# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for termsimport copy
import datetime
import os
import subprocess  # nosec
import unittest
from . import ScrewdriverTestCase
from tempfile import NamedTemporaryFile

from screwdrivercd.version.exceptions import VersionError
from screwdrivercd.version.version_types import versioners, Version, VersionGitRevisionCount, VersionSDV4Build, VersionDateSDV4Build, VersionUTCDate, VersionManualUpdate


existing_project_url_config = {
    'setup.cfg': b"""
[metadata]
author = Yahoo Python Platform Team
author_email = python@verizonmedia.com
name=mypyvalidator
project_urls = 
    Documentation = https://yahoo.github.io/mypyvalidator/
    Change Log = https://yahoo.github.io/python-screwdrivercd/changelog/
    CI Pipeline = https://cd.screwdriver.cd/pipelines/1
    Source = https://github.com/yahoo/mypyvalidator/tree/a5c3785ed8d6a35868bc169f07e40e889087fd2e
version=0.0.0

[options]
packages = find_namespace:
package_dir =
    =src

[options.packages.find]
where=src
"""
}


class TestVersioners(ScrewdriverTestCase):
    environ_keys = {
        'BASE_PYTHON', 'PACKAGE_DIR', 'PACKAGE_DIRECTORY', 'SD_ARTIFACTS_DIR', 'SD_BUILD', 'SD_BUILD_ID',
        'SD_PULL_REQUEST', 'SCM_URL', 'SD_BUILD_SHA',
    }

    def test__version__read_setup_version__no_version(self):
        version = Version(ignore_meta_version=True).read_setup_version()
        self.assertEqual(version, Version.default_version)

    def test__version__get_link_to_project_using_hash__unset_env_variables(self):
        link = Version(ignore_meta_version=True, link_to_project=True).get_link_to_project_using_hash()
        self.assertEqual(link, '')

    def test__version__get_link_to_project_using_hash__set_and_unset_env_variables(self):
        os.environ['SD_BUILD_SHA'] = 'a5c3785ed8d6a35868bc169f07e40e889087fd2e'
        link = Version(ignore_meta_version=True, link_to_project=True).get_link_to_project_using_hash()
        self.assertEqual(link, '')

    def test__version__get_link_to_project_using_hash__set_env_variables__https_git(self):
        os.environ['SCM_URL'] = 'https://github.com/org/project'
        os.environ['SD_BUILD_SHA'] = 'a5c3785ed8d6a35868bc169f07e40e889087fd2e'
        ver = Version(ignore_meta_version=True, link_to_project=True)
        link = ver.get_link_to_project_using_hash()
        self.assertEqual(link, 'https://github.com/org/project/tree/a5c3785ed8d6a35868bc169f07e40e889087fd2e')

        ver.update_setup_cfg_metadata()
        self.assertTrue(os.path.exists('setup.cfg'))
        with open('setup.cfg') as fh:
            setup_cfg = fh.read()
        self.assertIn(link, setup_cfg)

    def test__version__get_link_to_project_using_hash__set_env_variables__https_git__existing_project_urls(self):
        os.environ['SCM_URL'] = 'https://github.com/org/project'
        os.environ['SD_BUILD_SHA'] = 'a5c3785ed8d6a35868bc169f07e40e889087fd2e'
        self.write_config_files(existing_project_url_config)

        ver = Version(ignore_meta_version=True, link_to_project=True)
        link = ver.get_link_to_project_using_hash()
        self.assertEqual(link, 'https://github.com/org/project/tree/a5c3785ed8d6a35868bc169f07e40e889087fd2e')

        ver.update_setup_cfg_metadata()
        self.assertTrue(os.path.exists('setup.cfg'))
        with open('setup.cfg') as fh:
            setup_cfg = fh.read()
        self.assertIn(link, setup_cfg)
        self.assertIn('Documentation = https://yahoo.github.io/mypyvalidator/', setup_cfg)

    def test__version__get_link_to_project_using_hash__set_env_variables__ssh_git(self):
        os.environ['SCM_URL'] = 'git@github.com:org/project'
        os.environ['SD_BUILD_SHA'] = 'a5c3785ed8d6a35868bc169f07e40e889087fd2e'
        ver = Version(ignore_meta_version=True, link_to_project=True)
        link = ver.get_link_to_project_using_hash()
        self.assertEqual(link, 'https://github.com/org/project/tree/a5c3785ed8d6a35868bc169f07e40e889087fd2e')

        ver.update_setup_cfg_metadata()
        self.assertTrue(os.path.exists('setup.cfg'))
        with open('setup.cfg') as fh:
            setup_cfg = fh.read()
        self.assertIn(link, setup_cfg)

    def test__manual_version_update(self):
        with NamedTemporaryFile('w') as file:
            setup_cfg_content = '[metadata]\nversion = 1.0.5'
            file.write(setup_cfg_content)

            file.seek(0)
            version = VersionManualUpdate(setup_cfg_filename=file.name, ignore_meta_version=True)
            self.assertEqual(version.generate(), ['1', '0', '5'])

    def test__git_revision_count__no_git(self):
        with self.assertRaises(VersionError):
            version = str(VersionGitRevisionCount(ignore_meta_version=True, log_errors=False))

    def test__git_revision_count__git(self):
        try:
            self.setupEmptyGit()
        except subprocess.CalledProcessError:
            # Git cli not working
            return
        version = str(VersionGitRevisionCount(ignore_meta_version=True, log_errors=False))
        self.assertEqual(version, '0.0.1')
        with open('foo', 'w') as fh:
            fh.write('')
        subprocess.call(['git', 'add', 'foo'])
        subprocess.call(['git', 'commit', '-a', '-m', 'added foo'])
        version = str(VersionGitRevisionCount(ignore_meta_version=True, log_errors=False))
        self.assertEqual(version, '0.0.2')

    def test__git_revision_count__git_update(self):
        try:
            self.setupEmptyGit()
        except subprocess.CalledProcessError:
            # Git cli not working
            return
        versioner = VersionGitRevisionCount(ignore_meta_version=True, log_errors=False)
        versioner.update_setup_cfg_metadata()
        with open('setup.cfg') as setup_handle:
            result = setup_handle.read().strip()
        self.assertIn('version = 0.0.1', result)

    def test__sdv4_SD_BUILD__unset(self):
        self.delkeys(['SD_BUILD', 'SD_BUILD_ID', 'SD_PULL_REQUEST'])
        with self.assertRaises(VersionError):
            version = str(VersionSDV4Build(ignore_meta_version=True, log_errors=False))

    def test__sdv4_SD_BUILD__set(self):
        self.delkeys(['SD_BUILD', 'SD_BUILD_ID', 'SD_PULL_REQUEST'])
        os.environ['SD_BUILD'] = '9999'
        versioner = VersionSDV4Build(ignore_meta_version=True, log_errors=False)
        self.assertEqual(str(versioner), '0.0.9999')
        versioner.update_setup_cfg_metadata()
        config_version = Version().read_setup_version()
        self.assertEqual(config_version, ['0', '0', '9999'])

    def test__sdv4_SD_BUILD__PR__unset(self):
        self.delkeys(['SD_BUILD', 'SD_BUILD_ID', 'SD_PULL_REQUEST'])
        os.environ['SD_PULL_REQUEST'] = '1'
        with self.assertRaises(VersionError):
            version = str(VersionSDV4Build(ignore_meta_version=True, log_errors=False))

    def test__sdv4_SD_BUILD__PR__set(self):
        self.delkeys(['SD_BUILD', 'SD_BUILD_ID', 'SD_PULL_REQUEST'])
        os.environ['SD_BUILD'] = '9999'
        os.environ['SD_PULL_REQUEST'] = '9'
        versioner = VersionSDV4Build(ignore_meta_version=True, log_errors=False)
        self.assertEqual(str(versioner), '0.0.9999a9')
        versioner.update_setup_cfg_metadata()
        config_version = Version().read_setup_version()
        self.assertEqual(config_version, ['0', '0', '9999a9'])

    def test__utc_date(self):
        now = datetime.datetime.utcnow()
        expected = '{now.year}.{now.month}{now.day:02}.{now.hour:02}{now.minute:02}{now.second:02}'.format(now=now)
        version = str(VersionUTCDate(ignore_meta_version=True, now=now))
        self.assertEqual(version, expected)

    def test__sdv4_date(self):
        os.environ['SD_BUILD'] = '9999'
        now = datetime.datetime.utcnow()
        expected = f'{str(now.year)[-2:]}.{now.month}.{os.environ["SD_BUILD"]}'
        version = str(VersionDateSDV4Build(ignore_meta_version=True, now=now))
        self.assertEqual(version, expected)

    def test__sdv4_date__sd_build_id(self):
        os.environ['SD_BUILD_ID'] = '9999'
        now = datetime.datetime.utcnow()
        expected = f'{str(now.year)[-2:]}.{now.month}.{os.environ["SD_BUILD_ID"]}'
        version = str(VersionDateSDV4Build(ignore_meta_version=True, now=now))
        self.assertEqual(version, expected)

    def test__sdv4_date__pr(self):
        os.environ['SD_BUILD'] = '9999'
        os.environ['SD_PULL_REQUEST'] = '9'
        now = datetime.datetime.utcnow()
        expected = f'{str(now.year)[-2:]}.{now.month}.{os.environ["SD_BUILD"]}a{os.environ["SD_PULL_REQUEST"]}'
        version = str(VersionDateSDV4Build(ignore_meta_version=True, now=now))
        self.assertEqual(version, expected)

    def test__sdv4_date__unset(self):
        self.delkeys(['SD_BUILD', 'SD_BUILD_ID', 'SD_PULL_REQUEST'])
        with self.assertRaises(VersionError):
            version = str(VersionDateSDV4Build(ignore_meta_version=True))


if __name__ == '__main__':
    unittest.main()



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
screwdrivercd utility functions
"""
from .environment import env_bool, env_int, flush_terminals
from .screwdriver import create_artifact_directory


__all__ = ['contextmanagers', 'environment', 'exceptions', 'output', 'package', 'run', 'screwdriver', 'tox']



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Screwdriver github deploy key setup utility
"""
# The logging_basicConfig has to be run before other imports because some modules we use log output on import
# pylint: disable=wrong-import-order, wrong-import-position
from ..screwdriver.environment import logging_basicConfig
logging_basicConfig(check_prefix='GIT_DEPLOYKEY')

import base64
import distro
import logging
import os
import shutil
import subprocess  # nosec
import tempfile
from hashlib import sha256
from urllib.parse import urlparse

from ..installdeps.cli import main as installdeps_main
from ..utility.contextmanagers import InTemporaryDirectory

logger = logging.getLogger(__name__)


fingerprints = {
    'old github fingerprint': b'16:27:ac:a5:76:28:2d:36:63:1b:56:4d:eb:df:a6:48',     # Old github fingerprint, For openssh < 7.4
    'new github_fingerprint': b'SHA256:nThbg6kXUpJWGl7E1IGOCspRomTxdCARLviKw6E5SY8',  # New github fingerprint for openssh >= 7.4
}


ssh_agent_deploy_conf = """[build-system]
# Minimum requirements for the build system to execute.
requires = ["setuptools", "wheel"]  # PEP 508 specifications.

[tool.sdv4_installdeps]
    install = ['apk', 'apt-get', 'yum']

    [tool.sdv4_installdeps.apk]
        deps = ['openssh-client']
    
    [tool.sdv4_installdeps.apt-get]
        deps = ['openssh-client']
    
    [tool.sdv4_installdeps.yum]
        deps = ['openssh-clients']
"""


def git_key_secret() -> bytes:
    git_key = os.environ.get('GIT_DEPLOY_KEY', '')
    if not git_key:  # Nothing to do
        return b''

    git_key_decoded = base64.b64decode(git_key)
    if not git_key_decoded.startswith(b'-----BEGIN RSA PRIVATE KEY-----\n'):
        print('Decoded GIT_DEPLOY_KEY secret does not have a private key header')
    if not git_key_decoded.endswith(b'-----END RSA PRIVATE KEY-----\n'):
        print('Decoded GIT_DEPLOY_KEY secret does not have a private key footer')
    return git_key_decoded


def add_github_to_known_hosts(known_hosts_filename: str = '~/.ssh/known-hosts'):  # pragma: no cover
    """
    Add the github hosts to the known hosts

    Parameters
    ----------
    known_hosts_filename: str, optional
        The known_hosts file to update
    """
    known_hosts_filename = os.path.expanduser(known_hosts_filename)
    known_hosts_dirname = os.path.dirname(known_hosts_filename)
    os.makedirs(os.path.expanduser(known_hosts_dirname), exist_ok=True, mode=0o0700)
    os.chmod(os.path.dirname(known_hosts_filename), 0o0700)

    keyscan_command = shutil.which('ssh-keyscan')
    if not keyscan_command:
        keyscan_command = '/usr/bin/ssh-keyscan'

    with open(known_hosts_filename, 'ab') as fh:
        os.fchmod(fh.fileno(), 0o0600)
        fh.write(b'\n')
        subprocess.check_call([keyscan_command, 'github.com'], stdout=fh)  # nosec


def validate_known_good_hosts(known_hosts_filename: str = '~/.ssh/known-hosts') -> bool:  # pragma: no cover
    """
    Check the known hosts for the github hosts

    Returns
    -------
    bool
        True if at least one good host is present, False otherwise
    """
    known_hosts_filename = os.path.expanduser(known_hosts_filename)

    match = False
    output = subprocess.check_output(['ssh-keygen', '-l', '-f', known_hosts_filename])  # nosec
    for desc, fingerprint in fingerprints.items():
        if fingerprint not in output:
            print(f'Known github fingerprint {desc} is missing from known-hossts')
            continue
        match = True
    return match


def load_github_key(git_key):  # pragma: no cover
    """
    Load the github key into the ssh-agent
    """
    # subprocess.run(['ssh-add'], input=git_key)  # nosec
    # return
    git_key_passphrase = os.environ.get('GIT_DEPLOY_KEY_PASSPHRASE', '')
    with tempfile.TemporaryDirectory() as tempdir:
        os.makedirs(os.path.join(tempdir, '.ssh'), mode=0o0700)
        key_filename = os.path.join(tempdir, '.ssh/git_key')
        m = sha256()
        m.update(git_key)
        print(f'Private key hash: {m.hexdigest()}')
        with open(key_filename, 'wb') as fh:
            os.fchmod(fh.fileno(), 0o0600)
            fh.write(git_key)
        with open(f'{key_filename}.pub', 'wb') as fh:
            os.fchmod(fh.fileno(), 0o0644)
            command = ['ssh-keygen', '-vv', '-y', '-f', key_filename]
            command += ['-P', git_key_passphrase]
            subprocess.check_call(command, stdout=fh, timeout=15)  # nosec
        subprocess.check_call(['ssh-add', key_filename], stdin=subprocess.DEVNULL, timeout=15)  # nosec


def set_git_mail_config():  # pragma: no cover
    """
    Set the git mail config variables
    """
    subprocess.check_call(['git', 'config', '--global', 'user.email', "dev-null@screwdriver.cd"])  # nosec
    subprocess.check_call(['git', 'config', '--global', 'user.name', "sd-buildbot"])  # nosec


def update_git_remote():  # pragma: no cover
    """
    Update the git remote address to use the git protocol via ssh
    """
    new_git_url = None
    remote_output = subprocess.check_output(['git', 'remote', '-v'])  # nosec
    remote_output = remote_output.decode(errors='ignore')
    for line in remote_output.split('\n'):
        line = line.strip()
        splitline = line.split()
        if len(splitline) != 3:
            continue
        remote, old_git_url, remote_type = splitline
        if remote != 'origin':
            continue
        if remote_type != '(push)':
            continue
        if 'http' not in old_git_url:
            continue
        parsed_url = urlparse(old_git_url)
        new_git_url = f'git@{parsed_url.netloc}:{parsed_url.path.lstrip("/")}'
        break
    if new_git_url:
        subprocess.check_call(['git', 'remote', 'set-url', '--push', 'origin', new_git_url])  # nosec
        subprocess.check_call(['git', 'remote', 'set-url', 'origin', new_git_url])  # nosec
        subprocess.call(['git', 'remote', '-v'])  # nosec


def install_ssh_agent():  # pragma: no cover
    """
    Install ssh-agent if it is missing
    """
    missing = False
    for command in ['ssh-agent', 'ssh-keyscan']:
        if not shutil.which(command):  # Already installed
            missing = True
            break

    if not missing:
        return

    print('# Installing openssh client', flush=True)
    with InTemporaryDirectory():
        with open('pyproject.toml', 'w') as fh:
            fh.write(ssh_agent_deploy_conf)
        installdeps_main()


def setup_ssh_main() -> int:  # pragma: no cover
    """
    Github deploykey ssh setup, setup ssh eiddcchttjfenrtetcgbiglgcgfureejrluufcbjbngj
    so that ssh-agent can be startedx.

    Returns
    -------
    int:
        The returncode to be returned from the utility
    """
    git_key = git_key_secret()
    if not git_key:  # Nothing to do
        print('No GIT_DEPLOY_KEY secret present')
        return 0

    logger.debug('Installing ssh clients if it is not installed')
    install_ssh_agent()

    print('\n# Adding github.com to known_hosts')
    add_github_to_known_hosts()

    # The validation below does not work with the version of ssh-keygen on centos 5
    # so on that operating system we skip the validation.
    if distro.id() in ['rhel']:
        if distro.major_version() in ['5']:
            return 0

    print('\n# Validating known good hosts')
    validate_known_good_hosts()

    return 0


def add_deploykey_main() -> int:  # pragma: no cover
    """
    Github deploykey setup utility, this utility adds the keys from the screwdriver secrets into the ssh-agent.

    This tool requires that ssh-agent be running already.

    Returns
    -------
    int:
        The returncode to be returned from the utility
    """
    git_key = git_key_secret()
    if not git_key:  # Nothing to do
        return 0

    print('\n# Loading the github key into the ssh-agent')
    load_github_key(git_key)

    print('\n# Setting the git user.email and user.name config settings')
    set_git_mail_config()

    print('\n# Updating the git remote to use the ssh url')
    update_git_remote()

    add_github_to_known_hosts('/root/.ssh/known_hosts')

    return 0



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Type hint validation wrapper for screwdrivercd

This wrapper runs the validation tool.  This wrapper does the following:

- Runs the type checking tool with command line arguments to generate output in the formats specified by TYPE_CHECK_REPORT_FORMAT
- ensures all the logs and reports are stored in the build artifacts before exiting.
- Propagates a success code if the TYPE_CHECK_ENFORCING is set to false.
"""
# The logging_basicConfig has to be run before other imports because some modules we use log output on import
# pylint: disable=wrong-import-order, wrong-import-position
from ..screwdriver.environment import logging_basicConfig, update_job_status
logging_basicConfig(check_prefix='TEST_CHECK')

import logging
import os
import sys

from ..utility.output import header
from ..utility.tox import run_tox


logger_name = 'validate_test' if __name__ == '__main__' else __name__
logger = logging.getLogger(logger_name)


def main():
    tox_envlist = os.environ.get('TOX_ENVLIST', 'py38,py39,py310,py311')
    return run_tox(tox_envlist=tox_envlist)


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
screwdrivercd exceptions
"""


class PackageError(Exception):
    """
    General package error
    """


class PackageValidationError(PackageError):
    """
    Validation of the package failed
    """


class PackageParseError(PackageValidationError):
    """
    Parsing of the package configuration failed
    """

class TimeoutError(Exception):
    """A timeout occurred"""



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for termsimport copy
import os
from screwdrivercd.packaging.build_python import main as build_python_main
from screwdrivercd.packaging.publish_python import main as publish_python_main, package_exists, poll_until_available
from screwdrivercd.utility.environment import standard_directories
from vcr_unittest import VCRTestCase

from . import ScrewdriverTestCase


working_config = {
    'setup.py': b"""
from setuptools import setup
setup()
""",
    'setup.cfg': b"""
[metadata]
name=mypyvalidator
version=0.0.0

[options]
packages =
    mypyvalidator

package_dir =
    =src
""",
    'src/mypyvalidator/__init__.py': b"""a: int = 1\n"""
}


class ScrewdriverPackagingTestCase(ScrewdriverTestCase):

    def test__packaging__main__sdist(self):
        os.environ['PACKAGE_TYPES'] = 'sdist'
        self.write_config_files(working_config)
        build_python_main()

        print(os.listdir('artifacts/packages'))
        self.assertTrue(os.path.exists('artifacts/packages/mypyvalidator-0.0.0.tar.gz'))

    def test__packaging__main__wheel(self):
        os.environ['PACKAGE_TYPES'] = 'wheel'
        self.write_config_files(working_config)
        build_python_main()

        print(os.listdir('artifacts/packages'))
        self.assertTrue(os.path.exists('artifacts/packages/mypyvalidator-0.0.0-py3-none-any.whl'))


class ScrewdriverPackagingPublishTestCase(ScrewdriverTestCase):
    def moc_twine_command(self, rc=0, stdout='stdout', stderr='stderr'):
        twine_command_filename = os.path.join(self.tempdir.name, 'twine')
        with open(twine_command_filename, 'w') as fh:
            os.fchmod(fh.fileno(), 0o0755)
            fh.write(f'#!/bin/bash\nprintenv\n')
            if stdout:
                fh.write(f'\ncat << EOF\n{stdout}\nEOF\n')
            if stderr:
                fh.write(f'\ncat << EOF >&2\n{stderr}\nEOF\n')
            fh.write(f'exit {rc}\n')
        return twine_command_filename

    def test__publish__no_package_dir(self):
        os.environ['PUBLISH_PYTHON_TIMEOUT'] = "0"
        package_dir = standard_directories('publish_python')['packages']
        if os.path.exists(package_dir):
            print(package_dir)
            os.rename(package_dir, package_dir + '.disabled')
        rc = publish_python_main(twine_command=self.moc_twine_command(rc=999))
        self.assertEqual(rc, 0)

    def test__publish__disabled(self):
        os.environ['PUBLISH_PYTHON'] = 'False'
        os.environ['PUBLISH_PYTHON_TIMEOUT'] = "0"
        package_dir = standard_directories('publish_python')['packages']
        os.makedirs(package_dir, exist_ok=True)
        rc = publish_python_main(twine_command=self.moc_twine_command(rc=999))
        self.assertEqual(rc, 0)

    def test__publish__sdist__fails_user_secret(self):
        os.environ['PUBLISH_PYTHON_FAIL_MISSING_CRED'] = 'True'
        os.environ['PACKAGE_TYPES'] = 'sdist'
        os.environ['PUBLISH_PYTHON_TIMEOUT'] = "0"
        self.write_config_files(working_config)
        build_python_main()
        self.assertTrue(os.path.exists('artifacts/packages/mypyvalidator-0.0.0.tar.gz'))

        rc = publish_python_main(twine_command=self.moc_twine_command(rc=999))

        self.assertEqual(rc, 1)

    def test__publish__sdist__fails_user_password(self):
        os.environ['PUBLISH_PYTHON_FAIL_MISSING_CRED'] = 'True'
        os.environ['PYPI_USER'] = 'foo'
        os.environ['PACKAGE_TYPES'] = 'sdist'
        os.environ['PUBLISH_PYTHON_TIMEOUT'] = "0"
        self.write_config_files(working_config)
        build_python_main()
        self.assertTrue(os.path.exists('artifacts/packages/mypyvalidator-0.0.0.tar.gz'))

        rc = publish_python_main(twine_command=self.moc_twine_command(rc=999))

        self.assertEqual(rc, 1)

    def test__publish__sdist__fails_user_twine(self):
        os.environ['PUBLISH_PYTHON_FAIL_MISSING_CRED'] = 'True'
        os.environ['PYPI_USER'] = 'foo'
        os.environ['PYPI_PASSWORD'] = 'bar'
        os.environ['PACKAGE_TYPES'] = 'sdist'
        os.environ['PUBLISH_PYTHON_TIMEOUT'] = "0"
        self.write_config_files(working_config)
        build_python_main()
        self.assertTrue(os.path.exists('artifacts/packages/mypyvalidator-0.0.0.tar.gz'))

        rc = publish_python_main(twine_command=self.moc_twine_command(rc=999))

        self.assertEqual(rc, 1)

    def test__publish__sdist__pass__no_file_missing_cred(self):
        os.environ['PUBLISH_PYTHON_FAIL_MISSING_CRED'] = 'False'
        os.environ['PACKAGE_TYPES'] = 'sdist'
        os.environ['PUBLISH_PYTHON_TIMEOUT'] = "0"
        self.write_config_files(working_config)
        build_python_main()
        self.assertTrue(os.path.exists('artifacts/packages/mypyvalidator-0.0.0.tar.gz'))

        rc = publish_python_main(twine_command=self.moc_twine_command(rc=999))

        self.assertEqual(rc, 0)

    def test__publish__sdist(self):
        os.environ['PUBLISH_PYTHON_FAIL_MISSING_CRED'] = 'True'
        os.environ['PYPI_USER'] = 'foo'
        os.environ['PYPI_PASSWORD'] = 'bar'
        os.environ['PACKAGE_TYPES'] = 'sdist'
        os.environ['PUBLISH_PYTHON_TIMEOUT'] = "0"
        self.write_config_files(working_config)
        build_python_main()
        self.assertTrue(os.path.exists('artifacts/packages/mypyvalidator-0.0.0.tar.gz'))

        rc = publish_python_main(twine_command=self.moc_twine_command(rc=0))

        self.assertEqual(rc, 0)

    def test__publish__sdist__fail(self):
        os.environ['PUBLISH_PYTHON_FAIL_MISSING_CRED'] = 'True'
        os.environ['PYPI_USER'] = 'foo'
        os.environ['PYPI_PASSWORD'] = 'bar'
        os.environ['PACKAGE_TYPES'] = 'sdist'
        os.environ['PUBLISH_PYTHON_TIMEOUT'] = "0"
        self.write_config_files(working_config)
        build_python_main()
        self.assertTrue(os.path.exists('artifacts/packages/mypyvalidator-0.0.0.tar.gz'))

        rc = publish_python_main(twine_command=self.moc_twine_command(rc=999))

        self.assertEqual(rc, 1)

    def test__publish__sdist__test_pypi(self):
        os.environ['PUBLISH_PYTHON_FAIL_MISSING_CRED'] = 'True'
        os.environ['TEST_PYPI_USER'] = 'foo'
        os.environ['TEST_PYPI_PASSWORD'] = 'bar'
        os.environ['PACKAGE_TYPES'] = 'sdist'
        os.environ['TWINE_REPOSITORY_URL'] = 'https://test.pypi.org/legacy/'
        os.environ['PUBLISH_PYTHON_TIMEOUT'] = "0"
        self.write_config_files(working_config)
        build_python_main()
        self.assertTrue(os.path.exists('artifacts/packages/mypyvalidator-0.0.0.tar.gz'))

        rc = publish_python_main(twine_command=self.moc_twine_command(rc=0))

        self.assertEqual(rc, 0)


class ScrewdriverPackagingPollerTestCase(VCRTestCase):
    def _get_vcr_kwargs(self):
        """
        Set the VCR settings
        """
        return dict(
            record_mode='new_episodes',
            decode_compressed_response=True,
        )

    def test__package_exists__true(self):
        result = package_exists('setuptools', 'setuptools-68.0.0.tar.gz')
        self.assertTrue(result)

    def test__package_exists__false(self):
        result = package_exists('foozle', 'foozle-9999.9999.32.zip')
        self.assertFalse(result)

    def test__poll_until_available__ok(self):
        result = poll_until_available('serviceping', {'serviceping-19.5.78.tar.gz', 'serviceping-19.5.121522-py3-none-any.whl'}, timeout=30)
        self.assertIsInstance(result, set)
        self.assertFalse(result)

    def test__poll_until_available__fail(self):
        result = poll_until_available('serviceping', {'serviceping-19.5.78.tar.gz', 'serviceping-19.9999.121522-py3-none-any.whl'}, timeout=5)
        self.assertIsInstance(result, set)
        self.assertEqual(result, {'serviceping-19.9999.121522-py3-none-any.whl'})



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
screwdrivercd tox helpers
"""
import logging
import os
import shutil
import subprocess  # nosec


logger = logging.getLogger(__name__)


def store_tox_logs():
    """
    Store tox logs into the Screwdriver artifacts
    """
    artifacts_dir = os.environ.get('SD_ARTIFACTS_DIR', '')
    if not artifacts_dir:
        return

    if artifacts_dir and os.path.exists('.tox'):
        log_dir = os.path.join(artifacts_dir, 'logs/tox')
        os.makedirs(log_dir, exist_ok=True)
        for dirpath, dirnames, filenames in os.walk('.tox'):
            if len(dirpath) < 5:
                continue
            if dirpath.endswith('/log'):
                for filename in filenames:
                    source_filename = os.path.join(dirpath, filename)
                    dest_dir = os.path.join(log_dir, dirpath[5:])
                    dest_filename = os.path.join(dest_dir, filename)
                    os.makedirs(dest_dir, exist_ok=True)
                    shutil.copyfile(source_filename, dest_filename)


def run_tox(tox_envlist=None, tox_args=None):
    """
    Run tox
    """
    command = []

    # Use stdbuf to line buffer output if it is present
    if shutil.which('stdbuf'):  # pragma: no cover
        command += ['stdbuf', '-oL', '-eL']

    if shutil.which('pypirun'):  # pragma: no cover
        command += ['pypirun', 'tox']

    command += ['tox']

    if not tox_args:
        tox_args = os.environ.get('TOX_ARGS', '').split()
    if tox_envlist:
        command += ['-e', tox_envlist]
    if tox_args:
        command += tox_args

    # Run command
    rc = 0
    logger.debug(f'Running: {" ".join(command)}')
    try:
        subprocess.check_call(command)  # nosec
    except subprocess.CalledProcessError as error:
        rc = error.returncode

    store_tox_logs()

    return rc



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
import configparser
import copy
import os
import sys
from tempfile import TemporaryDirectory
import unittest
from screwdrivercd.version import cli
from screwdrivercd.version.version_types import versioners


class TestCLI(unittest.TestCase):
    cwd = None
    orig_argv = None
    orig_environ =None
    tempdir = None
    environ_keys = set('SD_BUILD')
    origcwd = os.getcwd()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.orig_argv = sys.argv
        self.cwd = os.getcwd()
        self.orig_environ = copy.copy(os.environ)

    def setUp(self):
        super().setUp()
        self.orig_argv = sys.argv
        self.tempdir = TemporaryDirectory()
        os.chdir(self.tempdir.name)

    def tearDown(self):
        super().tearDown()
        if self.orig_argv:
            sys.argv = self.orig_argv

        for environ_key in self.environ_keys:
            if self.orig_environ.get(environ_key, None):
                os.environ[environ_key] = self.orig_environ[environ_key]

        for versioner in versioners.values():
            versioner.ignore_meta_version = True
        os.chdir(self.origcwd)
        self.tempdir.cleanup()

    def _get_version(self, setup_cfg_filename='setup.cfg'):
        config = configparser.ConfigParser()
        config.read(setup_cfg_filename)
        if 'metadata' in config.sections():
            return config['metadata'].get('version', None)

    def test__main__no_setup_cfg(self):
        sys.argv = ['cli']
        cli.main()

    def test__main__setup_cfg__version(self):
        sys.argv = ['cli', '--ignore_meta']
        if 'SD_BUILD' in os.environ:
            del os.environ['SD_BUILD']
        print(f'current directory {os.getcwd()!r}')
        with open('setup.cfg', 'w') as fh:
            fh.write('[metadata]\nversion = 0.0.0\n[screwdrivercd.version]\nversion_type = git_revision_count\n')
        cli.main()
        with open('setup.cfg') as fh:
            result = fh.read()
        self.assertIn('version = 0.0.0', result)

    def test__main__setup_cfg__version__update_meta(self):
        sys.argv = ['cli', '--ignore_meta', '--update_meta']
        if 'SD_BUILD' in os.environ:
            del os.environ['SD_BUILD']
        print(f'current directory {os.getcwd()!r}')
        with open('setup.cfg', 'w') as fh:
            fh.write('[metadata]\nversion = 0.0.0\n[screwdrivercd.version]\nversion_type = git_revision_count\n')
        cli.main()
        with open('setup.cfg') as fh:
            result = fh.read()
        self.assertIn('version = 0.0.0', result)


if __name__ == '__main__':
    unittest.main()



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
screwdrivercd utility functions
"""
import os


def create_artifact_directory(artifact_directory: str = '') -> None:
    """
    Create the artifact directory if it is not present

    Parameters
    ----------
    artifact_directory: str, optional
        The artifact destination directory to create, if not specified will use the value of the SD_ARTIFACTS_DIR
        environment variable, if the variable is not present it will use a directory named `artifacts`
    """
    if not artifact_directory:
        artifact_directory = os.environ.get('SD_ARTIFACTS_DIR', 'artifacts')

    artifact_directory = str(artifact_directory)
    os.makedirs(artifact_directory, exist_ok=True)



#!/usr/bin/env python
# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
import os
import tempfile
import unittest.mock

from screwdrivercd.installdeps.installers.apk import ApkInstaller


CONFIG_FILE = 'pyproject.toml'
TEST_CONFIG = f'''[build-system]
# Minimum requirements for the build system to execute.
requires = ["setuptools", "wheel"]  # PEP 508 specifications.

[tool.sdv4_installdeps]
    install = ['apk']

    [tool.sdv4_installdeps.apk]
        deps = [
            'python3',
            'mysql-client'
        ]

'''


class TestApk(unittest.TestCase):
    original_environ = None

    def setUp(self):
        self._cwd = os.getcwd()
        super().setUp()
        self.original_environ = os.environ
        self.tempdir = tempfile.TemporaryDirectory()
        os.chdir(self.tempdir.name)
        with open(CONFIG_FILE, 'w') as config_handle:
            config_handle.write(TEST_CONFIG)

    def tearDown(self):
        super().tearDown()
        if self.original_environ:
            os.environ = self.original_environ
            self.original_environ = None
        os.chdir(self._cwd)
        self.tempdir.cleanup()

    @unittest.skipUnless(os.path.exists('/sbin/apk'), 'No apk binary present on system')
    def test__install__default(self):
        installer = ApkInstaller(dry_run=True)
        result = installer.install_dependencies()
        self.assertIn('python3', result)



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Screwdriver V4 System package dependency installer module.

This module contains modules to install operating system packages in a variety of formats.
"""
from typing import List


__all__: List[str] = ['cli', 'config', 'installer', 'installers', 'requirement']



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""Install yum dependencies"""
import logging
import os
import shutil
import subprocess  # nosec - All subprocess calls use full path
from typing import Dict, Optional, List

from termcolor import colored

from ..installer import Installer


LOG = logging.getLogger(__name__)


class YumInstaller(Installer):
    """
    Yum Package tool package installer
    """
    install_command: List[str] = ['yum', 'install', '-y']
    install_repo_command: List[str] = ['yum-config-manager', '--add-repo']
    config_section: str = 'yum'
    install_command_path: List[str] = ['/bin', '/usr/bin']
    use_system_path: bool = False
    supports_repositories: bool = True
    default_repos: Optional[Dict[str, str]] = dict()
    _repo_tool_install_failed = False

    def install_repo_tool(self):  # pragma: no cover - Function is OS specific
        """
        Installer the tool needed to add repositories to the system

        Raises
        ------
        FileNotFoundError: The tool was not found after attempting to install it
        """
        if self.install_repo_command[0].startswith('/') or self._repo_tool_install_failed:
            return

        install_command = shutil.which(self.install_repo_command[0])
        if install_command:
            self.install_repo_command[0] = install_command

        if not os.path.exists(self.install_repo_command[0]):
            LOG.debug(f'Tool to add repos, {self.install_repo_command[0]} is missing, trying to install it')
            subprocess.check_call([self.install_command[0], 'install', '-y', 'yum-utils'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)  # nosec - All subprocess calls use full path
            install_command = shutil.which(self.install_repo_command[0])
            if install_command:
                self.install_repo_command[0] = install_command
            else:
                self._repo_tool_install_failed = True
                raise FileNotFoundError('Could not find the {self.install_repo_command[0]} utility')

    def add_repo(self, repo_name, repo_url):  # pragma: no cover - Function is OS specific
        """
        Add Yum repos specified in the configuration
        """
        try:
            self.install_repo_tool()
        except FileNotFoundError:
            print(colored(f'The yum-config-manager utility is missing and cannot be installed, cannot add the {repo_name} repository', 'red'), flush=True)
            return

        if repo_url.startswith('enable:'):
            repo_name = repo_url[7:]
            if self.print_output:
                print(f'Enabling {repo_name!r} repo: ', end='', flush=True)
            try:
                subprocess.check_call([self.install_repo_command[0], '--enable', repo_name, '--save'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)  # nosec - All subprocess calls use full path
                if self.print_output:
                    print(colored('Ok', 'green'), flush=True)
            except subprocess.CalledProcessError:
                if self.print_output:
                    print(colored('Failed', 'red'), flush=True)
                else:
                    LOG.error('Enabling repository {repo_name!r} failed')
            return

        if repo_url.startswith('disable:'):
            repo_name = repo_url[8:]
            if self.print_output:
                print(f'Disabling {repo_name!r} repo: ', end='', flush=True)
            try:
                subprocess.check_call([self.install_repo_command[0], '--disable', repo_name, '--save'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)  # nosec - All subprocess calls use full path
                if self.print_output:
                    print(colored('Ok', 'green'), flush=True)
            except subprocess.CalledProcessError:
                if self.print_output:
                    print(colored('Failed', 'red'), flush=True)
                else:
                    LOG.error('Disabling repository {repo_name!r} failed')
            return

        if self.print_output:
            print(f'Adding {repo_name!r} with url {repo_url!r} repo: ', end='', flush=True)
        else:
            LOG.debug(f'Adding {repo_name!r} with url {repo_url!r} repo')
        try:
            subprocess.check_call(self.install_repo_command + [repo_url], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)  # nosec - All subprocess calls use full path
            if self.print_output:
                print(colored('Ok', 'green'), flush=True)
        except subprocess.CalledProcessError:
            if self.print_output:
                print(colored('Failed', 'red'), flush=True)
            else:
                LOG.error('Adding repository {repo_name!r} failed')

    def validate_dependency(self, dependency):  # pragma: no cover - Function is OS specific
        """
        Validate a dependency is valid
        Parameters
        ----------
        dependency: str
            The package name to validate
        Returns
        -------
        bool
            True if valid
        """
        try:
            subprocess.check_call([self.install_command[0], 'info', dependency], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)  # nosec - All subprocess calls use full path
        except subprocess.CalledProcessError:
            return False
        return True



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
screwdrivercd.documentation module
"""
from typing import List


__all__: List[str] = ['cli', 'exceptions', 'plugin', 'utility', 'mkdocs', 'sphinx']



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""Versioning plugin classes"""
import configparser
import logging
import os
import subprocess  # nosec
from datetime import datetime
from typing import List, Union
from .exceptions import VersionError


LOG = logging.getLogger(__name__)


class Version:
    """
    Base Screwdriver Versioning class
    """
    name: Union[str, None] = None
    default_version: List[str] = ['0', '0', '0']
    setup_cfg_filename: str = 'setup.cfg'
    _meta_version: str = ''

    def __init__(self, setup_cfg_filename=None, ignore_meta_version: bool = False, update_sdv4_meta: bool = True, meta_command: str = 'meta', link_to_project: bool = False):
        if setup_cfg_filename:  # pragma: no cover
            self.setup_cfg_filename = setup_cfg_filename
        self.meta_command = meta_command
        self.ignore_meta_version = ignore_meta_version
        self.update_sdv4_meta = update_sdv4_meta
        self.link_to_project = link_to_project

    def __repr__(self):
        return repr(self.version)

    def __str__(self):
        return self.version

    def commit_changed_setup_cfg(self):  # pragma: no cover
        """
        Git commit the setup.cfg
        """
        try:
            output = subprocess.check_output(['git', 'commit', '-m', 'Updated version', self.setup_cfg_filename])  # nosec
            LOG.debug(f'Git commit output {output}')
        except (FileNotFoundError, subprocess.CalledProcessError):
            pass

    def read_setup_version(self):
        """
        Read the package version from the setup.cfg file

        Returns
        -------
        str:
            The version number from the setup.cfg [metadata] section or the default version if the version is not
            present.
        """
        config = configparser.ConfigParser()
        config.read(self.setup_cfg_filename)
        if 'metadata' in config.sections():
            return config['metadata'].get('version', '').split('.')
        return self.default_version

    def generate(self):
        """
        Generate the version
        """
        return self.read_setup_version()

    def get_link_to_project_using_hash(self):
        """
        Generate and return link to build-triggering commit using its SHA hash
        """
        scm_url = os.environ.get('SCM_URL', '')
        sha = os.environ.get('SD_BUILD_SHA', '')
        if self.link_to_project and scm_url and sha:
            if scm_url.startswith('git@'):
                hostname = scm_url.split(':')[0].split('@')[1]
                path = ':'.join(scm_url.split(':')[1:])
                return f'https://{hostname}/{path}/tree/{sha}'
            else:
                return f'{scm_url}/tree/{sha}'
        return ''

    def update_setup_cfg_metadata(self):
        """
        Update the version value in the setup.cfg file
        """
        if not self.version:  # pragma: no cover
            return
        link_to_project = self.get_link_to_project_using_hash()
        config = configparser.ConfigParser()
        config.read(self.setup_cfg_filename)
        if 'metadata' not in config.sections():
            config['metadata'] = {}

        config['metadata']['version'] = self.version
        if link_to_project:
            project_urls_str = config['metadata'].get('project_urls', '')
            project_urls_dict = {}
            if project_urls_str:
                for entry in project_urls_str.split(os.linesep):
                    entry = entry.strip()
                    if '=' in entry:
                        key, value = entry.split('=')
                        key = key.strip()
                        value = value.strip()
                        project_urls_dict[key] = value

            project_urls_dict['Source'] = link_to_project

            project_urls_str = '\n'
            for key, value in project_urls_dict.items():
                project_urls_str += f'{key} = {value}\n'

            config['metadata']['project_urls'] = project_urls_str.rstrip('\n')

        with open(self.setup_cfg_filename, 'w') as config_file_handle:
            config.write(config_file_handle)

        self.commit_changed_setup_cfg()

    def update_meta_version(self):  # noqa
        """
        Update the meta_version value based on the generated version and pull_request_number
        """
        if not self.meta_version:  # pragma: no cover
            if not self.pull_request_number:
                new_version = self.generated_version
            else:  # pragma: no cover
                new_version = f'{self.generated_version}a{self.pull_request_number}'
            print(f'Updating the screwdriver metadata: package.version={new_version}', flush=True)
            self.meta_version = new_version

    @property
    def pull_request_number(self):
        """
        Return the Pull request number from the Screwdriver SD_PULL_REQUEST env variable if present, returns 0
        if it is not present.
        
        Returns
        -------
        int:
            Pull request number or 0 if not running from a pull request
        """
        try:
            prnum = int(os.environ.get('SD_PULL_REQUEST', None))
        except (TypeError, ValueError):
            prnum = 0
        return prnum

    @property
    def generated_version(self):
        """
        The generated version
        """
        return '.'.join(self.generate())

    @property
    def meta_version(self) -> str:
        """
        The version from the screwdriver metadata package.version value or None if not present.
        """
        if self.ignore_meta_version:
            return ''
        if self._meta_version:  # pragma: no cover
            return self._meta_version
        try:  # pragma: no cover
            self._meta_version = subprocess.check_output([self.meta_command, 'get', 'package.version']).decode(errors='ignore').strip()  # nosec
        except (FileNotFoundError, subprocess.CalledProcessError):  # pragma: no cover
            pass
        if self._meta_version == 'null':
            self._meta_version = ''
        return self._meta_version  # pragma: no cover

    @meta_version.setter
    def meta_version(self, new_version):
        if not self.update_sdv4_meta:  # pragma: no cover
            return
        try:
            subprocess.check_call([self.meta_command, 'set', 'package.version', new_version])  # nosec
        except FileNotFoundError:  # pragma: no cover
            LOG.warning('The screwdriver meta command is missing, unable to set version in screwdriver metadata')
        self._meta_version = new_version

    @property
    def version(self):
        """
        The version from the versioner, will be the version from the screwdriver package.version if it is set, otherwise
        it will be the generated version number from the versioner.
        """
        if self.meta_version:  # pragma: no cover
            return self.meta_version

        ver = self.generated_version
        if not self.pull_request_number:
            return ver
        return f'{ver}a{self.pull_request_number}'


class VersionManualUpdate(Version):
    """
    Version updater that just reads the setup.cfg and takes the `version`
    number from `metadata` section. Be aware when using this updater, as this
    updater does not add any unique identifier to the version number.

    It is thus the caller's responsibility to ensure uniqueness is maintained
    during version update.
    """
    name: str = 'manual_update'


class VersionUpdateRevision(Version):
    """
    Version updater that updates the revision (last component) of a semetic version.
    """
    log_errors: bool = True
    default_revision_value: str = '0'
    
    def __init__(self, *args, **kwargs):
        self.log_errors = kwargs.pop('log_errors', self.log_errors)
        super().__init__(*args, **kwargs)

    def revision_value(self):  # pragma: no cover
        """Method to return a newly generated revision value"""
        return self.default_revision_value

    def generate(self):  # pragma: no cover
        version = super().generate()
        if version:
            try:
                version[-1] = self.revision_value()
            except VersionError as error:
                if self.log_errors:
                    LOG.exception(f'Unable to get revision value for versioner {self.name!r}')
                else:
                    raise error
        return version


class VersionGitRevisionCount(VersionUpdateRevision):
    """
    Version Revision updater that sets the revision number to be equal to the number of git commits of the current
    repository.
    
    Each new git commit will increment the revision value.
    
    Notes
    -----
    This Versioner may not work correctly if the current git repository is a shallow git clone.
    """
    name: str = 'git_revision_count'

    def revision_value(self):
        """
        Get a count of the number of git commits and return the count
        Returns
        -------
        int
            Revision count or None if git was not found
        """
        try:
            result = subprocess.check_output(['git', 'rev-list', '--count', 'HEAD']).decode(errors='ignore')  # nosec
        except subprocess.CalledProcessError:  # pragma: no cover
            try:  # pragma: no cover
                result = len(subprocess.check_output(['git', 'rev-list', 'HEAD']).decode(errors='ignore').split('\n'))  # nosec
            except subprocess.CalledProcessError:
                raise VersionError('Unable to generate a version from the git revision count')
        try:
            result = int(result)
        except ValueError:  # pragma: no cover
            raise VersionError('Got invalid response from the git rev-list command')
        return str(result)


class VersionSDV4Build(VersionUpdateRevision):
    """
    Version Revision updater that sets the revision number to be equal to the value of the screwdriver SD_BUILD number.
    
    Each new screwdriver job run will increment the revision number.
    """
    name = 'sdv4_SD_BUILD'

    def revision_value(self):
        revision = os.environ.get('SD_BUILD', None)
        if not revision:
            LOG.debug('No value for SD_BUILD found')
            revision = os.environ.get('SD_BUILD_ID', None)
        if not revision:
            raise VersionError('Unable to generate version, no SD_BUILD or SD_BUILD_ID value set in the environment variables')
        return revision


class VersionUTCDate(Version):
    """
    Version updater that generates a version based on the current UTC date/time.
    
    Each new screwdriver job will get a version based on the current date and time.
    """
    name = 'utc_date'
    now = None

    def __init__(self, *args, **kwargs):
        self.now = kwargs.pop('now', None)
        super().__init__(*args, **kwargs)

    def generate(self):
        now = datetime.utcnow()
        if self.now:  # pragma: no cover
            now = self.now
        return [f'{now.year}', f'{now.month}{now.day:02}', f'{now.hour:02}{now.minute:02}{now.second:02}']


class VersionDateSDV4Build(Version):
    """
    Version updater that generates a version based on the current year and month and the screwdriver build ID
    """
    name = 'sdv4_date'

    def __init__(self, *args, **kwargs):
        self.now = kwargs.pop('now', None)
        super().__init__(*args, **kwargs)

    def generate(self):
        now = self.now if self.now else datetime.utcnow()
        revision = os.environ.get('SD_BUILD', None)
        if not revision:
            revision = os.environ.get('SD_BUILD_ID', None)
        if not revision:
            raise VersionError('Unable to generate version, no SD_BUILD or SD_BUILD_ID value set in the environment variables')
        return [f'{str(now.year)[-2:]}', f'{now.month}', revision]


versioners = {
    'default': VersionGitRevisionCount,
    VersionGitRevisionCount.name: VersionGitRevisionCount,
    VersionUTCDate.name: VersionUTCDate,
    VersionSDV4Build.name: VersionSDV4Build,
    VersionDateSDV4Build.name: VersionDateSDV4Build,
    VersionManualUpdate.name: VersionManualUpdate
}

# Make sure the versioners are listed all lowercase to make identifying them easier
for key, value in list(versioners.items()):
    if key.lower() not in versioners.keys():
        versioners[key.lower()] = value



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for termsimport copy
import copy
import os
import sys
import unittest

from json import dumps

from screwdrivercd.packaging.build_python import build_sdist_package, build_wheel_packages
from screwdrivercd.utility.environment import standard_directories
from screwdrivercd.utility.output import header
from screwdrivercd.utility.tox import run_tox
from screwdrivercd.validation.validate_dependencies import validate_with_safety
from screwdrivercd.validation.validate_package_quality import validate_package_quality
from screwdrivercd.validation.validate_style import main as style_main
from screwdrivercd.validation.validate_style import validate_codestyle
from screwdrivercd.validation.validate_type import main as type_main
from screwdrivercd.validation.validate_type import validate_type
from screwdrivercd.validation.validate_unittest import main as unittest_main

from . import ScrewdriverTestCase


class ScrewdriverValidatorTestCase(ScrewdriverTestCase):
    """
    Base class to test validation wrappers
    """


# Simple working package tree
working_config = {
    'setup.py': b"""
from setuptools import setup
setup()
""",
    'setup.cfg': b"""
[metadata]
author = Verizon Media Python Platform Team
author_email = python@verizonmedia.com
name=mypyvalidator
url=https://foo.bar.com/
version=0.0.0

[options]
packages =
    mypyvalidator

package_dir =
    =src
""",
    'src/mypyvalidator/__init__.py': b"""a: int = 1\n""",
    'tests/test.py': b"""from unittest import TestCase

class DummyTestCase(TestCase):
    def test_dummy(self):
        import mypyvalidator
        self.assertEqual(mypyvalidator.a, 1)
""",
    'tox.ini': b"""
[config]
package_dir = src/mypyvalidator
package_name = mypyvalidator

[tox]
envlist = py39,py310,py311
skip_missing_interpreters = true

[testenv]
changedir = {toxinidir}
commands = 
     python -c "import sys;print(sys.version_info)"
"""
}

# No source package tree
working_config_nosrc = copy.deepcopy(working_config)
working_config_nosrc['mypyvalidator/__init__.py'] = working_config_nosrc['src/mypyvalidator/__init__.py']
del working_config_nosrc['src/mypyvalidator/__init__.py']
working_config_nosrc['setup.cfg'] = b"""
[metadata]
name=mypyvalidator
version=0.0.0

[options]
packages =
    mypyvalidator
"""

# src dir has different case than package name
working_config_case_different = copy.deepcopy(working_config)
working_config_case_different['src/MyPyvalidator/__init__.py'] = working_config['src/mypyvalidator/__init__.py']
del working_config_case_different['src/mypyvalidator/__init__.py']
working_config_case_different['setup.cfg'].replace(b'\n    mypyvalidator', b'\n    MyPyvalidator')

# No source dir different case
working_config_case_different_nosrc = copy.deepcopy(working_config_case_different)
working_config_case_different_nosrc['MyPyvalidator/__init__.py'] = working_config_case_different_nosrc['src/MyPyvalidator/__init__.py']
del working_config_case_different_nosrc['src/MyPyvalidator/__init__.py']

# Invalid package tree
invalid_type_config = copy.deepcopy(working_config)
invalid_type_config['src/mypyvalidator/__init__.py'] = b"""a: int='1'"""

# Insecure dependency
insecure_dep_config = copy.deepcopy(working_config)
insecure_dep_config['setup.cfg'] = b"""
[metadata]
name=mypyvalidator
version=0.0.0

[options]
install_requires =
    insecure-package
    
packages =
    mypyvalidator

package_dir =
    =src
"""

# tox fail
tox_fail_config = copy.deepcopy(working_config)
tox_fail_config['tox.ini'] = b"""
[config]
package_dir = src/mypyvalidator
package_name = mypyvalidator

[tox]
envlist = py39,py310,py311
skip_missing_interpreters = true

[testenv]
changedir = {toxinidir}
commands = 
	python -c "import sys;sys.exit(1)"
"""

invalid_style_config = copy.deepcopy(working_config)
invalid_style_config['src/mypyvalidator/__init__.py'] = b"""a: int =1\n"""

class DepValidatorTestcase(ScrewdriverTestCase):
    validator_name = 'type_validation'

    @unittest.skip("No longer works")
    def test__secure_deps(self):
        self.write_config_files(working_config)
        result = validate_with_safety()
        self.assertEqual(result, 0)

    def test__insecure_deps(self):
        self.write_config_files(insecure_dep_config)
        artifacts_dir = os.environ.get('SD_ARTIFACTS_DIR', '')
        report_dir = os.path.join(artifacts_dir, 'reports/dependency_validation')
        json_report_filename = os.path.join(report_dir, 'safetydb.json')
        result = validate_with_safety()
        self.assertGreater(result, 0)
        self.assertTrue(os.path.exists(json_report_filename))


class StyleValidator(ScrewdriverTestCase):
    validator_name = 'style_validation'

    def test__style__pass__srcdir_main(self):
        os.environ['PACKAGE_DIR'] = 'src'
        self.write_config_files(working_config)
        result = style_main()
        self.assertEqual(result, 0)

    def test__style__pass__srcdir(self):
        os.environ['PACKAGE_DIR'] = 'src'
        self.write_config_files(working_config)
        result = validate_codestyle()
        self.assertEqual(result, 0)

    def test__style__pass__srcdir__args(self):
        os.environ['PACKAGE_DIR'] = 'src'
        os.environ['CODESTYLE_ARGS'] = '--show-source'
        self.write_config_files(working_config)
        result = validate_codestyle()
        self.assertEqual(result, 0)

    def test__style__fail__srcdir(self):
        os.environ['PACKAGE_DIR'] = 'src'
        self.write_config_files(invalid_style_config)
        result = validate_codestyle()
        self.assertGreater(result, 0)

    def test__style__pass__nosrc(self):
        self.write_config_files(working_config_nosrc)
        result = validate_codestyle()
        self.assertEqual(result, 0)

    def test_style__pass__srcdir_wrong_case(self):
        os.environ['PACKAGE_DIR'] = 'src'
        self.write_config_files(working_config_case_different)
        os.system('ls -lR')
        result = validate_codestyle()
        self.assertEqual(result, 0)

    def test_style__pass__srcdir_wrong_case__missing(self):
        os.environ['PACKAGE_DIR'] = 'srcz'
        self.write_config_files(working_config_case_different)
        os.system('ls -lR')
        result = validate_codestyle()
        self.assertNotEqual(result, 0)

    def test_style__pass__nosrcdir_wrong_case(self):
        self.write_config_files(working_config_case_different_nosrc)
        os.system('ls -lR')
        os.system('cat setup.cfg')
        result = validate_codestyle()
        self.assertEqual(result, 1)


class TypeValidatorTestcase(ScrewdriverTestCase):
    validator_name = 'type_validation'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.environ_keys.add('MYPY_ARGS')
        self.environ_keys.add('TYPE_CHECK_ENFORCING')
        self.environ_keys.add('TYPE_CHECK_REPORT_FORMAT')

    def test__main__valid(self):
        os.environ['PACKAGE_DIR'] = 'src'
        os.environ['TYPE_CHECK_ENFORCING'] = 'True'
        self.write_config_files(working_config)
        result = type_main()
        self.assertEqual(result, 0)

    def test__main__valid_nosrc(self):
        os.environ['TYPE_CHECK_ENFORCING'] = 'True'
        self.write_config_files(working_config_nosrc)
        result = type_main()
        self.assertEqual(result, 0)

    def test__mypy__valid(self):
        os.environ['PACKAGE_DIR'] = 'src'
        os.environ['TYPE_CHECK_ENFORCING'] = 'True'
        self.write_config_files(working_config)
        result = validate_type()
        self.assertEqual(result, 0)

    def test__mypy__valid__nopackage_dir__env(self):
        os.environ['TYPE_CHECK_ENFORCING'] = 'True'
        self.write_config_files(working_config)
        result = validate_type()
        self.assertEqual(result, 0)

    def test__mypy__valid__mypy_args(self):
        os.environ['PACKAGE_DIR'] = 'src'
        os.environ['TYPE_CHECK_ENFORCING'] = 'True'
        os.environ['MYPY_ARGS'] = '--show-error-context'

        self.write_config_files(working_config)
        result = validate_type()
        self.assertEqual(result, 0)

    def test__mypy__invalid_not_enforcing(self):
        os.environ['PACKAGE_DIR'] = 'src'
        os.environ['TYPE_CHECK_ENFORCING'] = 'False'

        self.write_config_files(invalid_type_config)
        result = validate_type()
        self.assertEqual(result, 0)

    def test__mypy__invalid_enforcing(self):
        os.environ['PACKAGE_DIR'] = 'src'
        os.environ['TYPE_CHECK_ENFORCING'] = 'True'

        self.write_config_files(invalid_type_config)
        result = validate_type()
        self.assertGreater(result, 0)
        self.assertTrue(os.path.exists(f'{self.artifacts_dir}/reports/{self.validator_name}/index.txt'))

    def test__mypy__invalid_enforcing_lxml_report(self):
        os.environ['PACKAGE_DIR'] = 'src'
        os.environ['TYPE_CHECK_ENFORCING'] = 'True'
        os.environ['TYPE_CHECK_REPORT_FORMAT'] = 'junit-xml'
        print(dumps(dict(os.environ), indent=4, sort_keys=True))
        self.write_config_files(invalid_type_config)
        result = validate_type()
        self.assertGreater(result, 0)
        self.assertTrue(os.path.exists(f'{self.artifacts_dir}/reports/{self.validator_name}/mypy.xml'))


class PackageQualityValidatorTestCase(ScrewdriverTestCase):
    validator_name = 'package_quality_validator'

    def test__quality__no_package_directory(self):
        package_dir = standard_directories('package_quality_validation')['packages']
        if os.path.exists(package_dir):
            os.rename(package_dir, package_dir + '.disabled')
        result = validate_package_quality()
        self.assertEqual(result, 0)

    def test__quality__no_package_default_fail(self):
        result = validate_package_quality()
        self.assertEqual(result, 0)

    def test__quality__no_package_fail(self):
        os.environ['VALIDATE_PACKAGE_QUALITY_FAIL_MISSING'] = 'True'
        result = validate_package_quality()
        self.assertEqual(result, 1)

    def test__quality__no_package_nofail(self):
        os.environ['VALIDATE_PACKAGE_QUALITY_FAIL_MISSING'] = 'False'
        result = validate_package_quality()
        self.assertEqual(result, 0)

    def test__quality__fail(self):
        os.environ['PYROMA_MIN_SCORE'] = '9'
        self.write_config_files(working_config)
        build_sdist_package()
        result = validate_package_quality()
        self.assertNotEqual(result, 0)

    def test__quality__pass__0(self):
        os.environ['PYROMA_MIN_SCORE'] = '0'
        self.write_config_files(working_config)
        build_sdist_package()
        result = validate_package_quality()
        self.assertEqual(result, 0)

    def test__quality__skip_wheel_pass__0(self):
        os.environ['PYROMA_MIN_SCORE'] = '0'
        self.write_config_files(working_config)
        build_sdist_package()
        build_wheel_packages()
        result = validate_package_quality()
        self.assertEqual(result, 0)


class UnitTestValidatorTestCase(ScrewdriverTestCase):
    validator_name = 'unittest'

    @unittest.skip("No longer works with newer tox releases")
    def test__unittest__default(self):
        self.write_config_files(working_config)
        result = unittest_main()
        for dirpath, dirs, files in os.walk('.tox'):
            for fname in files:
                if fname.endswith('.log'):
                    filename = os.path.join(dirpath, fname)
                    header(filename)
                    with open(filename) as fh:
                        print(fh.read() + os.linesep)
        self.assertEqual(result, 0)

    @unittest.skip("No longer works with newer tox releases")
    def test__unittest__fail(self):
        self.write_config_files(tox_fail_config)
        result = unittest_main()
        for dirpath, dirs, files in os.walk('.tox'):
            for fname in files:
                if fname.endswith('.log'):
                    filename = os.path.join(dirpath, fname)
                    header(filename)
                    with open(filename) as fh:
                        print(fh.read() + os.linesep)
        self.assertNotEqual(result, 0)

    @unittest.skip("No longer works with newer tox releases")
    def test__unittest__no_artifacts_dir(self):
        del os.environ['SD_ARTIFACTS_DIR']
        self.write_config_files(working_config)
        result = unittest_main()
        for dirpath, dirs, files in os.walk('.tox'):
            for fname in files:
                if fname.endswith('.log'):
                    filename = os.path.join(dirpath, fname)
                    header(filename)
                    with open(filename) as fh:
                        print(fh.read() + os.linesep)
        self.assertEqual(result, 0)

    @unittest.skip("No longer works with newer tox releases")
    def test__unittest__tox_args(self):
        os.environ['TOX_ARGS'] = '-v'
        self.write_config_files(working_config)
        result = unittest_main()
        for dirpath, dirs, files in os.walk('.tox'):
            for fname in files:
                if fname.endswith('.log'):
                    filename = os.path.join(dirpath, fname)
                    header(filename)
                    with open(filename) as fh:
                        print(fh.read() + os.linesep)
        self.assertEqual(result, 0)

    @unittest.skip("No longer works with newer tox releases")
    def test__unittest__tox_args_arg(self):
        self.write_config_files(working_config)
        result = run_tox()
        for dirpath, dirs, files in os.walk('.tox'):
            for fname in files:
                if fname.endswith('.log'):
                    filename = os.path.join(dirpath, fname)
                    header(filename)
                    with open(filename) as fh:
                        print(fh.read() + os.linesep)
        self.assertEqual(result, 0)

    @unittest.skip("No longer works with newer tox releases")
    def test__unittest__tox_envlist(self):
        os.environ['TOX_ENVLIST'] = f'py{sys.version_info.major}{sys.version_info.minor}'
        self.write_config_files(working_config)
        result = unittest_main()
        for dirpath, dirs, files in os.walk('.tox'):
            for fname in files:
                if fname.endswith('.log'):
                    filename = os.path.join(dirpath, fname)
                    header(filename)
                    with open(filename) as fh:
                        print(fh.read() + os.linesep)
        self.assertEqual(result, 0)

    @unittest.skip("No longer works with newer tox releases")
    def test__unittest__tox_envlist_arg(self):
        self.write_config_files(working_config)
        result = run_tox(tox_envlist=f'py{sys.version_info.major}{sys.version_info.minor}')
        for dirpath, dirs, files in os.walk('.tox'):
            for fname in files:
                if fname.endswith('.log'):
                    filename = os.path.join(dirpath, fname)
                    header(filename)
                    with open(filename) as fh:
                        print(fh.read() + os.linesep)
        self.assertEqual(result, 0)



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Dependency security validation wrapper for screwdrivercd

This wrapper runs the validation tool.  This wrapper does the following:

- Runs the dependency checking tool
- ensures all the logs and reports are stored in the build artifacts before exiting.
"""
# The logging_basicConfig has to be run before other imports because some modules we use log output on import
# pylint: disable=wrong-import-order, wrong-import-position
from ..screwdriver.environment import logging_basicConfig, update_job_status
logging_basicConfig(check_prefix='DEPENDENCY_CHECK')

import json
import logging
import os
import sys

from termcolor import colored
from pypirun.cli import install_and_run, interpreter_parent
from ..utility import create_artifact_directory


logger_name = 'validate_dependencies' if __name__ == '__main__' else __name__
logger = logging.getLogger(logger_name)


def validate_with_safety():
    """Run the safety command in a virtualenv to validate the package dependencies"""

    artifacts_dir = os.environ.get('SD_ARTIFACTS_DIR', '')
    report_dir = os.path.join(artifacts_dir, 'reports/dependency_validation')
    create_artifact_directory(report_dir)
    
    interpreter = interpreter_parent(sys.executable)
    interpreter = os.environ.get('BASE_PYTHON', interpreter)

    # Generate a full text report
    full_report_filename = os.path.join(report_dir, 'safetydb.full')
    rc = install_and_run(package='safety<2.0.0,.', command=f'safety check --full-report -o "{full_report_filename}"', interpreter=interpreter, upgrade_setuptools=True, upgrade_pip=True)

    if rc == 0:
        # update_job_status(status='SUCCESS', message='Dependency check passed')
        print(colored('Safetydb check passed', color='green'))
        return rc

    # Generate the report in json format
    json_report_filename = os.path.join(report_dir, 'safetydb.json')
    rc = install_and_run(package='safety<2.0.0,.', command=f'safety check --json -o "{json_report_filename}"', interpreter=interpreter, upgrade_setuptools=True, upgrade_pip=True)

    bad_packages = -1
    if os.path.exists(json_report_filename):  # pragma: no cover
        with open(json_report_filename) as fh:
            results = json.load(fh)
            bad_packages = len(results)

    update_job_status(status='FAILURE', message=f'Dependency check failed {bad_packages} bad dependencies')

    return rc


def main():  # pragma: no cover
    """
    Install dependencies utility command line entry point

    Returns
    -------
    int:
        Returncode from running the check
    """
    return validate_with_safety()


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Screwdriver metadata management functions
"""
import json
import logging
import subprocess#  # nosec


class Metadata(dict):
    """
    Screwdriver metadata wrapper

    all attributes to this class that are accessed or set will case the screwdriver meta command to run with the
    get or set for that specified value.

    Any underscore characters in the attribute name will be converted to period prior to getting/setting the value
    from screwdriver's metadata.
    """
    read_only = False

    def __getattr__(self, key):  # pragma: no cover
        """Get value from screwdriver metadata"""
        key = key.replace('_', '.')
        return self.get(key)

    def __setattr__(self, key, value):  # pragma: no cover
        """Set an attr to screwdriver metadata"""
        key = key.replace('_', '.')
        self.set(key, value)

    def get(self, key):  # pragma: no cover
        """
        Get a value from the screwdriver metadata store
        """
        command = ['meta', 'get', '--json-value', key]
        logging.debug(f'Running: {command}')
        try:
            response = subprocess.check_output(command).decode(errors='ignore')  # nosec
        except (subprocess.CalledProcessError, FileNotFoundError):
            return None
        result = json.loads(response)
        return result

    def set(self, key, value):  # pragma: no cover
        """
        Set a value in the screwdriver metadata store
        """
        if self.read_only:
            logging.debug('Cannot write metadata, metadata is readonly')
            return
        command = ['meta', 'set', '--json-value', key, json.dumps(value)]
        logging.debug(f'Running: {command}')
        try:
            subprocess.check_call(command)  # nosec
        except (subprocess.CalledProcessError, FileNotFoundError):
            return



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
screwdrivercd.documentation module command line utility
"""
# The logging_basicConfig has to be run before other imports because some modules we use log output on import
# pylint: disable=wrong-import-order, wrong-import-position
from ..screwdriver.environment import logging_basicConfig, update_job_status
logging_basicConfig(check_prefix='DOCUMENTATION')
import logging
import os

from termcolor import colored

from .exceptions import DocBuildError, DocPublishError
from .plugin import build_documentation, publish_documentation
from ..utility.environment import env_bool, is_pull_request


logger_name = __name__
if logger_name == '__main__':  # pragma: no cover
    logger_name = os.path.basename(__file__)
logger = logging.getLogger(logger_name)


def main():  # pragma: no cover
    """
    screwdrivercd.documentation cli main entrypoint

    Returns
    -------
    int, optional:
        returncode - Returncode from the publication operation, 0=success
    """
    documentation_formats = os.environ.get('DOCUMENTATION_FORMATS', 'mkdocs,sphinx')

    if documentation_formats:
        documentation_formats = [_.strip() for _ in documentation_formats.split(',')]

    documentation_publish_default = False if is_pull_request() else True

    rc = 0
    if env_bool('DOCUMENTATION_PUBLISH', documentation_publish_default):  # pragma: no cover
        operation = 'Published'
        try:
            publish_documentation(documentation_formats=documentation_formats)
        except DocPublishError:
            rc = 1
        if is_pull_request():
            print(colored('Warning: Published documentation from a pull request', 'yellow'))
    else:
        operation = 'Generated'
        try:
            build_documentation(documentation_formats=documentation_formats)
        except DocBuildError:
            rc = 1
    if is_pull_request():
        status = 'SUCCESS' if rc == 0 else 'FAILURE'
        # update_job_status(status=status, message=f'{operation} {", ".join(documentation_formats)} documentation')
    return rc



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""Install pip dependencies"""
import logging
import os
from typing import Optional, List

from ..installer import Installer


LOG = logging.getLogger(__name__)


class PipInstaller(Installer):
    """
    Python pip3 package installer
    """
    install_command: List[str] = ['pip3', 'install']
    config_section: str = 'pip3'
    bin_dir: Optional[str] = None
    install_command_path: Optional[List[str]] = [
        # Python platform python
        '/opt/python/bin',

        # Manylinux python
        '/opt/python/cp37-cp37m/bin',
        '/opt/python/cp36-cp36m/bin',
        '/opt/python/cp35-cp35m/bin',

        # Ubuntu/Fedora
        '/usr/bin'
    ]

    def find_install_command(self):
        """
        Find the install command binary to use and update the install command
        """
        base_python = os.environ.get('BASE_PYTHON', '')
        if base_python and os.path.exists(base_python):
            self.install_command = [base_python, '-m', 'pip', 'install']
            return
        super().find_install_command()



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Read/Parse the configuration file for the screwdrivercd.installdeps tool
"""
import copy
import os

from collections.abc import Mapping  # pylint: disable no-name-in-module
from typing import Any, Dict, Optional

import toml


CONFIGURATION_SCHEMA = {
    'apk': {
        'deps': [],
    },
    'apt-get': {
        'deps': [],
        'repos': {}
    },
    'echo': {
        'deps': []
    },
    'install': ['apk', 'apt-get', 'yinst', 'yum', 'pip3'],
    'pip3': {
        'deps': [],
        'repos': {}
    },
    'yinst': {
        'deps': [],
        'deps_current': [],
        'deps_quarantine': [],
        'deps_stable': [],
        'deps_test': []
    },
    'yum': {
        'deps': [],
        'branch_deps': {},
        'repos': {},
        'setting': {},
    }
}


def deep_update(source, update_dict):
    """
    Merge update_dict values into the source dictionary.
    """
    for key, value in update_dict.items():
        if isinstance(value, Mapping) and value:
            returned = deep_update(source.get(key, {}), value)
            source[key] = returned
        else:
            source[key] = update_dict[key]
    return source


class Configuration():
    """
    installdeps configuration class
    """
    filename = 'pyproject.toml'
    configuration: Dict[Any, Any] = {}

    def __init__(self, filename: Optional[str] = None):
        if filename:
            self.filename = filename
        self.configuration = copy.deepcopy(CONFIGURATION_SCHEMA)
        self.load_configuration()

    def load_configuration(self):
        """
        Load the configuration from the configuration file
        """
        if not os.path.exists(self.filename):
            return
        
        with open(self.filename) as fh:
            conf = toml.load(fh)

        tool_configs = conf.get('tool', None)
        if not tool_configs:
            return

        sdv4_installdeps_configs = tool_configs.get('sdv4_installdeps', None)
        if not sdv4_installdeps_configs:
            sdv4_installdeps_configs = tool_configs.get('screwdrivercd_installdeps', None)
            if not sdv4_installdeps_configs:
                return

        deep_update(self.configuration, sdv4_installdeps_configs)



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
import os
import subprocess
import sys
import tempfile
import time

from datetime import timedelta
from pathlib import Path
from pprint import pprint

from screwdrivercd.screwdriver.environment import update_job_status
from screwdrivercd.utility.contextmanagers import InTemporaryDirectory, Timeout, revert_file, working_dir
from screwdrivercd.utility.environment import env_bool, env_int, flush_terminals, ins_filename, interpreter_bin_command, standard_directories
from screwdrivercd.utility.exceptions import TimeoutError
from screwdrivercd.utility.output import header, print_error, status_message
from screwdrivercd.utility.package import run_setup_command, setup_query, PackageMetadata
from screwdrivercd.utility.screwdriver import create_artifact_directory
from screwdrivercd.utility.run import run_and_log_output
from screwdrivercd.utility.tox import run_tox, store_tox_logs

from . import ScrewdriverTestCase


tox_stub = {
    '.tox/log/GLOB-0.log': b'glob\n',
    '.tox/py38/log/py38-0.log': b'foo\n',
}


class TestPlatformTestUtility(ScrewdriverTestCase):

    def test_working_dir(self):
        """
        Verify the workingdir context manager switches to the directory
        and switches back correctly.
        :return:
        """

        cwd = os.getcwd()
        tempdir = tempfile.mkdtemp()

        # Change into the temp directory and get the cwd value, we can't use
        # the path for tempdir because the values could be different if
        # there are symlinks in the path.

        os.chdir(tempdir)
        expected_result = os.getcwd()
        os.chdir(cwd)

        with working_dir(expected_result):
            self.assertEqual(os.getcwd(), expected_result)
        self.assertEqual(os.getcwd(), cwd)
        os.removedirs(tempdir)

    def test_working_dir_nested(self):
        with tempfile.TemporaryDirectory() as tempdir:
            cwd = os.getcwd()

            # Change into the tempdir and use getcwd() so that our asserts
            # work properly when the temp directory has a symlink.
            os.chdir(tempdir)
            real_tempdir = os.getcwd()

            dir1 = os.path.join(os.getcwd(), '1')
            dir2 = os.path.join(os.getcwd(), '2')
            os.mkdir(dir1)
            os.mkdir(dir2)
            with working_dir(dir1):
                self.assertEqual(dir1, os.getcwd())
                with working_dir(dir2):
                    self.assertEqual(dir2, os.getcwd())
                self.assertEqual(dir1, os.getcwd())
            self.assertEqual(real_tempdir, os.getcwd())
            os.chdir(cwd)

    def test_revert_file__exists(self):
        with InTemporaryDirectory() as tempdir:
            tfile = os.path.join(tempdir, 'junk')
            with open(tfile, 'w') as fh:
                fh.write('Junk file\n')
            with revert_file(tfile):
                with open(tfile, 'w') as fh:
                    fh.write('Different\n')
                self.assertIn('Different', open(tfile).read())
            self.assertIn('Junk file', open(tfile).read())

    def test_revert_file__missing(self):
        with InTemporaryDirectory() as tempdir:
            tfile = os.path.join(tempdir, 'junk')
            with revert_file(tfile):
                with open(tfile, 'w') as fh:
                    fh.write('Different\n')
                self.assertIn('Different', open(tfile).read())
            self.assertFalse(os.path.exists(tfile))

    def test__InTemporaryDirectory(self):
        cwd = os.getcwd()
        with InTemporaryDirectory() as tempdir:
            self.assertNotEqual(cwd, tempdir)
            self.assertTrue(os.path.isdir(tempdir))
        self.assertFalse(os.path.exists(tempdir))
        self.assertEqual(cwd, os.getcwd())

    def test__Timeout__alarm_timeout(self):
        with self.assertRaises(TimeoutError):
            with Timeout(timedelta(seconds=1), use_alarm=True):
                time.sleep(2)

    def test__Timeout__noalarm_timeout(self):
        with self.assertRaises(TimeoutError):
            with Timeout(timedelta(milliseconds=10), use_alarm=False):
                time.sleep(2)

    def test__Timeout__default(self):
        with Timeout():
            time.sleep(.1)

    def test__flush_terminals(self):
        flush_terminals()

    def test__interpreter_bin_command__default(self):
        result = interpreter_bin_command()
        self.assertIsInstance(result, (str))
        self.assertIn('/python', result)
        self.assertTrue(os.path.exists(result))

    def test__interpreter_bin_command__command_exists(self):
        result = interpreter_bin_command('screwdrivercd_version')
        self.assertIsInstance(result, (str))
        self.assertTrue(result.endswith('/screwdrivercd_version'))
        self.assertTrue(os.path.exists(result))

    def test__interpreter_bin_command__command_missing__default(self):
        result = interpreter_bin_command('screwdrivercd_test__interpreter_bin_command__command_missing')
        self.assertIsInstance(result, (str))
        self.assertEqual(result, 'screwdrivercd_test__interpreter_bin_command__command_missing')

    def test__interpreter_bin_command__command_missing__fallback_path__false(self):
        result = interpreter_bin_command('screwdrivercd_test__interpreter_bin_command__command_missing', fallback_path=False)
        self.assertIsInstance(result, (str))
        self.assertFalse(result)  # Empty string evaluates as False

    def test__environement__env_int(self):
        os.environ['test__environement__env_int'] = '1'
        result = env_int('test__environement__env_int')
        del os.environ['test__environement__env_int']
        self.assertEqual(result, 1)

    def test__environement__env_int__default(self):
        result = env_int('test__environement__env_int__default')
        self.assertEqual(result, 0)

    def test_environment_env_bool_default_true(self):
        self.assertTrue(env_bool('TEST_UTILITY_ENV_BOOL', True))

    def test_environment_env_bool_default_false(self):
        self.assertFalse(env_bool('TEST_UTILITY_ENV_BOOL', False))

    def test_environment_env_bool_true(self):
        os.environ['TEST_UTILITY_ENV_BOOL'] = 'True'
        self.assertTrue(env_bool('TEST_UTILITY_ENV_BOOL', False))

    def test_environment_env_bool_false(self):
        os.environ['TEST_UTILITY_ENV_BOOL'] = 'False'
        self.assertFalse(env_bool('TEST_UTILITY_ENV_BOOL', True))

    def test__create_artifact_directory_env(self):
        os.environ['SD_ARTIFACTS_DIR'] = f'{self.tempdir.name}/test_artifacts'
        self.assertFalse(os.path.exists(os.environ['SD_ARTIFACTS_DIR']))
        create_artifact_directory()
        self.assertTrue(os.path.exists(os.environ['SD_ARTIFACTS_DIR']))

    def test__create_artifact_directory_passed(self):
        self.assertFalse(os.path.exists(f'{self.tempdir.name}/test_artifacts'))
        create_artifact_directory(f'{self.tempdir.name}/test_artifacts')
        self.assertTrue(os.path.exists(f'{self.tempdir.name}/test_artifacts'))

    def test__ins_filename__nodir(self):
        Path('foo').mkdir()
        result = ins_filename('foo')
        self.assertEqual(result, 'foo')

    def test__ins_filename__nodir__differentcase(self):
        Path('Goo').mkdir()
        result = ins_filename('goo')
        self.assertEqual(result, 'Goo')

    def test__ins_filename__fullpath(self):
        Path('foo').mkdir()
        fullpath = os.path.join(self.tempdir.name, 'foo')
        result = ins_filename(fullpath)
        self.assertEqual(result, fullpath)

    def test__ins_filename__dirnoexist(self):
        result = ins_filename('foo')
        self.assertEqual(result, '')

    def test__run__run_log_output__success(self):
        with InTemporaryDirectory():
            result = run_and_log_output(['echo', 'hello'], 'echo.log')
            self.assertTrue(os.path.exists('echo.log'))
            with open('echo.log') as fh:
                log_output = fh.read()
            self.assertEqual(log_output, 'hello\n')

    def test__run__run_log_output__fail(self):
        testscript_content = """import sys
print('hello')
sys.exit(1)
"""
        with InTemporaryDirectory():
            with open('testscript.py', 'w') as fh:
                fh.write(testscript_content)
            with self.assertRaises(subprocess.CalledProcessError):
                result = run_and_log_output([sys.executable, 'testscript.py'], 'test.log')
            self.assertTrue(os.path.exists('test.log'))
            with open('test.log') as fh:
                test_output = fh.read()
            self.assertEqual(test_output, 'hello\n')

    def test__setup_query(self):
        with InTemporaryDirectory():
            with open('setup.py', 'w') as setup_py_handle:
                setup_py_handle.write('from setuptools import setup\nsetup(name="foo")\n')
            result = setup_query('--name')
            self.assertEqual(result, 'foo')

    def test__setup_query__cli_args(self):
        with InTemporaryDirectory():
            with open('setup.py', 'w') as setup_py_handle:
                setup_py_handle.write('from setuptools import setup\nsetup(name="foo")\n')
            result = setup_query('--name', cli_args='--verbose')
            self.assertEqual(result, 'foo')

    def test__PackageMetadata__archive(self):
        with InTemporaryDirectory():
            with open('setup.py', 'w') as setup_py_handle:
                setup_py_handle.write('from setuptools import setup\nsetup(name="foo", version="0.0.0")\n')
            setup_query('sdist')
            expected_filename = f'dist/foo-0.0.0.tar.gz'
            self.assertTrue(os.path.exists(expected_filename))
            PackageMetadata(expected_filename)

    def test__PackageMetadata__archive_zip(self):
        with InTemporaryDirectory():
            with open('setup.py', 'w') as setup_py_handle:
                setup_py_handle.write('from setuptools import setup\nsetup(name="foo", version="0.0.0")\n')
            setup_query('sdist', cli_args=['--formats', 'zip'])
            expected_filename = f'dist/foo-0.0.0.zip'
            self.assertTrue(os.path.exists(expected_filename))
            PackageMetadata(expected_filename)

    def test__update_job_status__invalid_status(self):
        with self.assertRaises(KeyError):
            update_job_status('BADDD', message='mojo')

    def test__standard_directories__no_command(self):
        os.environ['SD_ARTIFACTS_DIR'] = self.tempdir.name
        result = standard_directories()
        self.assertEqual(result['artifacts'], f'{self.tempdir.name}')
        self.assertEqual(result['documentation'], f'{self.tempdir.name}/documentation')
        self.assertEqual(result['logs'], f'{self.tempdir.name}/logs')
        self.assertEqual(result['packages'], f'{self.tempdir.name}/packages')
        self.assertEqual(result['reports'], f'{self.tempdir.name}/reports')
        for value in result.values():
            self.assertTrue(os.path.exists(value))

    def test__standard_directories__command(self):
        os.environ['SD_ARTIFACTS_DIR'] = self.tempdir.name
        result = standard_directories(command='foo')
        pprint(result)
        self.assertEqual(result['artifacts'], f'{self.tempdir.name}')
        self.assertEqual(result['documentation'], f'{self.tempdir.name}/documentation')
        self.assertEqual(result['logs'], f'{self.tempdir.name}/logs{os.sep}foo')
        self.assertEqual(result['packages'], f'{self.tempdir.name}/packages')
        self.assertEqual(result['reports'], f'{self.tempdir.name}/reports{os.sep}foo')
        for value in result.values():
            self.assertTrue(os.path.exists(value))

    def test__header(self):
        header('This is a header message')

    def test__header__stderr(self):
        header('This is a header message', outfile=sys.stderr)

    def test__header__collapse(self):
        header('This is a header message', collapse=True)

    def test__status_message__default(self):
        result = status_message('Cool')
        self.assertEqual(result, 'Cool')

    def test__print_error(self):
        print_error('Bad operator at keyboard')

    def test_store_tox_logs(self):
        self.write_config_files(tox_stub)
        store_tox_logs()
        self.assertTrue(os.path.exists(f'{self.artifacts_dir}/logs/tox/log/GLOB-0.log'))
        self.assertTrue(os.path.exists(f'{self.artifacts_dir}/logs/tox/py38/log/py38-0.log'))

    def test_store_tox_logs_no_artifacts_dir(self):
        del os.environ['SD_ARTIFACTS_DIR']
        self.write_config_files(tox_stub)
        store_tox_logs()
        self.assertFalse(os.path.exists('artifacts/logs'))
        self.assertFalse(os.path.exists('artifacts/logs/tox/log/GLOB-0.log'))
        self.assertFalse(os.path.exists('artifacts/logs/tox/py38/log/py38-0.log'))



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""Command line argument parsing"""
import argparse
import configparser
from .exceptions import VersionError
from .version_types import versioners


def get_config_default(key, default=None, setup_cfg_filename='setup.cfg'):
    """
    Get the default value for a configuration setting

    Parameters
    ==========
    key: str
        The key to read from the configuration

    default: str, optional
        A default value to return if the key is not present in the configuration.

    setup_cfg_filename: str
        The configuration file to parse for the configuration value.
    """
    config = configparser.ConfigParser()
    config.read(setup_cfg_filename)
    if 'screwdrivercd.version' in config.sections():
        return config['screwdrivercd.version'].get(key, default)

    if 'sdv4.version' in config.sections():
        return config['sdv4.version'].get(key, default)

    if 'ouroath.platform_version' in config.sections():  # pragma: no cover
        return config['ouroath.platform_version'].get(key, default)

    return default


def get_bool_equivalent(key) -> bool:
    """ Get the equivalent bool value for a string key in the config file. """

    if isinstance(key, str) and key.lower() in ['false', '0', 'off']:
        return False
    return True


def parse_arguments():
    """
    Parse the command line arguments

    Returns
    -------
    argparse.arguments:
        The parsed arguments
    """
    version_type = get_config_default('version_type', default='default')
    update_meta = get_config_default('update_screwdriver_meta', default='false')
    update_meta = get_bool_equivalent(update_meta)
    link_to_project = get_config_default('link_to_project', default='false')
    link_to_project = get_bool_equivalent(link_to_project)

    version_choices = list(versioners.keys())
    if version_type not in version_choices:
        raise VersionError(f'The version_type in the [screwdrivercd.version] section of setup.cfg has an invalid version type of {version_type!r}')

    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--force_update', default=False, action='store_true', help='Update the setup.cfg even if it does not have a metadata section')
    parser.add_argument('--version_type', default=version_type, choices=version_choices, help='Type of version number to generate')
    parser.add_argument('--ignore_meta', default=False, action='store_true', help='Ignore the screwdriver v4 metadata')
    parser.add_argument('--update_meta', default=update_meta, action='store_true', help='Update the screwdriver v4 metadata with the new version')
    parser.add_argument('--link_to_project', default=link_to_project, action='store_true', help='Add/update link to source project tree for current package version')
    result = parser.parse_args()
    return result



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Python packaging command line wrapper
"""
import sys
from .build_python import main


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
import os
import tempfile
import unittest.mock
import subprocess
import sys

import distro
from pypirun.cli import interpreter_parent
from screwdrivercd.installdeps.installers.pip3 import PipInstaller
from screwdrivercd.utility.contextmanagers import InTemporaryDirectory


CONFIG_FILE = 'pyproject.toml'
TEST_CONFIG = f'''[build-system]
# Minimum requirements for the build system to execute.
requires = ["setuptools", "wheel"]  # PEP 508 specifications.

[tool.sdv4_installdeps]
    install = ['apk', 'apt-get', 'yinst', 'yum', 'pip3']

    [tool.sdv4_installdeps.pip3]
        deps = [
            'serviceping',
            'safety;distro_version=="{distro.version()}"', 'pypirun;distro_version!="{distro.version()}"'
        ]
'''


class TestPep3(unittest.TestCase):
    original_environ = None

    def setUp(self):
        self._cwd = os.getcwd()
        super().setUp()
        self.original_environ = os.environ
        self.tempdir = tempfile.TemporaryDirectory()
        os.chdir(self.tempdir.name)
        with open(CONFIG_FILE, 'w') as config_handle:
            config_handle.write(TEST_CONFIG)
        self.venv_dir = os.path.join(self.tempdir.name, 'venv')
        self.venv_bin_dir = os.path.join(self.venv_dir, 'bin')
        interpreter = interpreter_parent(sys.executable)
        subprocess.check_call([interpreter, '-m', 'venv', self.venv_dir])

    def tearDown(self):
        super().tearDown()
        if self.original_environ:
            os.environ = self.original_environ
            self.original_environ = None
        os.chdir(self._cwd)
        self.tempdir.cleanup()

    def test__install__default(self):
        installer = PipInstaller(bin_dir=self.venv_bin_dir)
        result = installer.install_dependencies()
        os.system(f'ls -lR {os.path.join(self.tempdir.name, "venv")}/bin')
        self.assertListEqual(result, ['serviceping', 'safety'])
        expected_command = os.path.join(self.tempdir.name, 'venv/bin/serviceping')
        self.assertTrue(os.path.exists(expected_command), f'Command not found {expected_command!r}')
        expected_command = os.path.join(self.tempdir.name, 'venv/bin/safety')
        self.assertTrue(os.path.exists(expected_command), f'Command not found {expected_command!r}')
        expected_command = os.path.join(self.tempdir.name, 'venv/bin/pypirun')
        self.assertFalse(os.path.exists(expected_command), f'Command found {expected_command!r}')




# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Screwdriver environment functions
"""
import json
import logging
import os
import shutil
from ..utility.environment import env_bool, is_pull_request
from .metadata import Metadata


def get_env_job_name(default='') -> str:  # pragma: no cover
    """
    Get the job name from the screwdriver environment that matches the name as reported to github

    Returns
    -------
    str:
        Job name, will be an empty string if a name is not present in the environment
    """
    job_name = os.environ.get('SD_JOB_NAME', default).split(':')[-1]
    return job_name


def logging_basicConfig(**kwargs):
    """
    Call logging.basicConfig() with default values based on the screwdriver environment variables
    """
    level = kwargs.pop('level', None)
    check_prefix = kwargs.pop('check_prefix', '')
    if check_prefix and not level:
        kwargs['level'] = logging.DEBUG if env_bool(f'{check_prefix}_DEBUG', default=False) else logging.INFO
    return logging.basicConfig(**kwargs)


def update_job_status(status='SUCCESS', message='Everything looks good!'):  # pragma: no cover
    """
    Update the job status in the screwdriver metadata

    Parameters
    ----------
    status: str, optional
        Must be SUCCESS or FAILURE

    message: str, optional
        The status message to use

    Raises
    ------
    KeyError: Status is not SUCCESS or FAILURE
    """
    if status not in ['SUCCESS', 'FAILURE']:
        raise KeyError(f'Status {status} not found in ["SUCCESS", "FAILURE"]')

    if not shutil.which('meta'):
        return

    metadata = Metadata()
    metadata.set(f'meta.status.{get_env_job_name()}', json.dumps(dict(status=status, message=message)))



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Python packaging command line wrapper
"""
import os
import shlex
import subprocess  # nosec
import sys

from typing import List

from ..utility.package import setup_query
from ..utility.run import run_and_log_output


def build_sdist_package():
    """
    Build a Python sdist format package

    Parameters
    ----------
    build_log_dir
    package_artifacts

    Returns
    -------
    str
        Path to the created package or None if creation failed
    """
    artifacts_dir = os.environ.get('SD_ARTIFACTS_DIR', 'artifacts')
    package_artifacts = os.path.join(artifacts_dir, 'packages')
    build_log_dir = os.path.join(artifacts_dir, 'logs/build')
    setup_args = shlex.split(os.environ.get('SETUP_ARGS', ''))

    before = []
    if os.path.exists('dist'):
        before = os.listdir('dist')

    run_and_log_output(command=[sys.executable, 'setup.py', 'sdist'] + setup_args, logfile=f'{build_log_dir}/sdist_build.log')
    if os.path.exists('dist'):
        for filename in os.listdir('dist'):
            if filename in before:  # pragma: no cover
                continue
            source_filename = f'dist/{filename}'
            dest_filename = f'{package_artifacts}/{filename}'
            os.makedirs(package_artifacts, exist_ok=True)
            print(f'Moving {source_filename} -> {dest_filename}')
            os.rename(source_filename, dest_filename)
            return [dest_filename], []
    return [], []


def build_wheel_packages():
    artifacts_dir = os.environ.get('SD_ARTIFACTS_DIR', 'artifacts')
    manylinux = os.environ.get('MANYLINUX', 'True').lower() in ['1', 'true', 'on']
    plat = os.environ.get('AUDITWHEEL_PLAT', '')
    package_artifacts = os.path.join(artifacts_dir, 'packages')
    build_log_dir = os.path.join(artifacts_dir, 'logs/build')
    setup_args = shlex.split(os.environ.get('SETUP_ARGS', ''))
    wheel_build_dir = os.path.join(artifacts_dir, 'wheelbuild')

    # Get values from the setup.py/setup.cfg
    package_name = setup_query('--name')

    print('# Generating wheel package', flush=True)
    failed = set()
    built_packages = set()
    before = set(os.listdir('dist'))
    run_and_log_output([sys.executable, 'setup.py', 'bdist_wheel'] + setup_args, logfile=f'{build_log_dir}/wheel_build.log')
    after = set(os.listdir('dist'))
    for filename in after - before:
        if filename.endswith('none-any.whl'):
            print('Package is generating a pure python wheel, skipping manylinux packaging', flush=True)
            dest_filename = f'{package_artifacts}/{filename}'
            print(f'Moving dist/{filename} -> {dest_filename}')
            os.rename(f'dist/{filename}', f'{dest_filename}')
            return [dest_filename], []

    if manylinux and plat:  # pragma: no cover
        print('\n# Generating manylinux wheels', flush=True)
        for entry in os.listdir('/opt/python'):
            interpreter_dir = os.path.join('/opt/python', entry)
            pip_command = f'{interpreter_dir}/bin/pip'
            if os.path.isdir(interpreter_dir) and os.path.exists(pip_command):
                print(f'Generating wheel for the {entry!r} python interpreter', flush=True)
                try:
                    run_and_log_output([pip_command, 'wheel', '.', '-w', wheel_build_dir], logfile=f'{build_log_dir}/wheel_build_{entry}.log')
                except subprocess.CalledProcessError:
                    failed.add(entry)

        print('\n## Bundling shared libraries into the wheel packages', flush=True)
        for wheel in os.listdir(wheel_build_dir):
            full_wheel_filename = os.path.join(wheel_build_dir, wheel)
            if not wheel.startswith(package_name + '-'):
                # print(f'Removing dependency wheel {wheel!r}', flush=True)
                os.remove(full_wheel_filename)
                continue
            if wheel.endswith('none-any.whl'):
                # Not a binary wheel
                continue
            print(f'Bundling shared libraries for wheel package {wheel!r}', flush=True)

            # This has to be run in a separate virtualenv using pypirun because auditwheel currently pins the
            # version of the wheel package which creates package dependency conflicts.
            run_and_log_output(
                ['pypirun', 'auditwheel', 'auditwheel', 'repair', full_wheel_filename, '--plat', plat, '-w', wheel_build_dir],
                logfile=f'{build_log_dir}/wheel_auditwheel_{wheel}.log'
            )
            os.remove(full_wheel_filename)

    print('\n# Storing generated packages in the build artifacts')
    for filename in os.listdir(wheel_build_dir):
        dest_filename = f'{package_artifacts}/{filename}'
        print(f'Moving {wheel_build_dir}/{filename} -> {dest_filename}', flush=True)
        os.rename(f'{wheel_build_dir}/{filename}', dest_filename)
        built_packages.add(dest_filename)

    if failed:
        print(f'Package build failed for {failed}')
    return list(built_packages), failed


def main():
    # Get settings from the environment
    artifacts_dir = os.environ.get('SD_ARTIFACTS_DIR', 'artifacts')
    package_artifacts = os.path.join(artifacts_dir, 'packages')
    wheel_build_dir = os.path.join(artifacts_dir, 'wheelbuild')
    build_log_dir = os.path.join(artifacts_dir, 'logs/build')
    package_types = [_.strip().lower() for _ in os.environ.get('PACKAGE_TYPES', 'sdist,wheel').split(',')]

    os.makedirs(package_artifacts, exist_ok=True)
    os.makedirs(build_log_dir, exist_ok=True)
    os.makedirs(wheel_build_dir, exist_ok=True)
    os.makedirs('dist', exist_ok=True)

    built_packages = failed_packages = []
    for package_type in package_types:
        built: List[str] = []
        failed: List[str] = []
        if package_type == 'sdist':
            print(f'# Building sdist package', flush=True)
            built, failed = build_sdist_package()
        elif package_type == 'wheel':
            print('# Building wheel package(s)', flush=True)
            built, failed = build_wheel_packages()
        if built:
            built_packages += built
            print('    Created packages:')
            for package in built:
                print(f'        {package}', flush=True)
        if failed:  # pragma: no cover
            failed_packages += failed
            print('    Failed to create packages:')
            for package in failed:
                print(f'        {package}', flush=True)
    return len(failed_packages)


if __name__ == '__main__':
    main()



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Screwdriver Utility Runtime environment functions
"""
import logging
import os
import sys
from typing import Dict, Union

from pypirun.cli import interpreter_parent


logger = logging.getLogger(__name__)


def env_bool(variable_name: str, default: Union[None, bool] = None) -> Union[None, bool]:
    """
    Return the value of an env variable as a boolean

    Parameters
    ----------
    variable_name: str
        The environment variable

    default: bool, optional
        The default value to return if the env variable is not present.  Default=None

    Returns
    -------
    bool or None:
        The value from the env variable if present, otherwise the default value.

    """
    true_strings = ['true', 'on', '1']

    if variable_name not in os.environ.keys():
        return default

    value = os.environ.get(variable_name, str(default)).lower()

    if value in true_strings:
        return True

    return False


def env_int(variable_name: str, default: int = 0) -> int:
    """
    Return the value of an environment variable as an integer

    Parameters
    ----------
    variable_name: str
        Name of the environment variable

    default: int
        Default value to return if the environment variable is not present

    Returns
    -------
    int:
        Value of the env variable as an integer, or the default value if the env variable is not present.
    """
    if variable_name not in os.environ.keys():
        return default

    value = os.environ.get(variable_name, str(default)).lower()
    return int(value)


def flush_terminals():
    """
    Flush the terminal devices
    """
    sys.stdout.flush()
    sys.stderr.flush()


def interpreter_bin_command(command: str='', fallback_path: bool=True) -> str:
    """
    Return the full path to a command in the current interpreter's bin directory.

    This command handles finding the bin directory, even if the interpreter is running within a virtualenv

    Parameters
    ----------
    command: str, optional
        The command to find in the interpreter bin directory, default=python

    Returns
    -------
    str:
        Full path to the command.  If the command is not present returns command if fallback_path is True or an empty
        string otherwise.
    """
    if not command:
        command = os.path.basename(sys.executable)
    bin_dir = os.path.dirname(sys.executable)
    new_command = os.path.join(bin_dir, command)
    if os.path.exists(new_command):
        return new_command
    bin_dir = os.path.dirname(interpreter_parent(sys.executable))
    new_command = os.path.join(bin_dir, command)
    if os.path.exists(new_command):  # pragma: no cover
        return new_command
    if fallback_path:
        return command
    return ''


def standard_directories(command: str='') -> Dict[str, str]:
    """
    Dictionary of standard directory locations for a command

    Parameters
    ----------
    command: str, optional
        The name of the command to get the directories for, if no command is given the base directory will be used

    Returns
    -------
    dict of str
        A dictionary of directory types and paths
    """
    sep = ''
    if command:
        sep = os.sep

    artifacts_dir: str = os.environ.get('SD_ARTIFACTS_DIR', 'artifacts')
    directories: Dict[str, str] = {
        'artifacts': artifacts_dir,
        'documentation': os.path.join(artifacts_dir, f'documentation'),
        'logs': os.path.join(artifacts_dir, f'logs{sep}{command}'),
        'packages': os.path.join(artifacts_dir, f'packages'),
        'reports': os.path.join(artifacts_dir, f'reports{sep}{command}'),
    }
    for directory in directories.values():
        if directory:  # pragma: no cover
            os.makedirs(directory, exist_ok=True)

    return directories


def ins_filename(filename: str) -> str:
    """
    Return the filename of an actual file or directory that case insensitively resolves to the filename given.  If no
    matching files or directories exists, returns an empty string.
    """
    contain_dir = os.path.dirname(filename)
    base_name = os.path.basename(filename).lower()
    if not contain_dir:
        contain_dir = '.'
    if not os.path.exists(contain_dir):
        return ''
    for f in os.listdir(contain_dir):
        if base_name == f.lower():
            if os.path.dirname(filename):
                return os.path.join(os.path.dirname(filename), f)
            else:
                return f
    return ''


def is_pull_request() -> bool:
    """
    Return True if the SD_PULL_REQUEST environment variable has a PULL Request value, False otherwise
    """
    pr_num_str = os.environ.get('SD_PULL_REQUEST', '')
    if not pr_num_str:
        return False

    try:
        pr_num = int(pr_num_str)
    except ValueError:
        return False

    return True



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
import os
import pathlib
import tempfile
import unittest

from screwdrivercd.documentation.utility import clean_directory, copy_contents
from screwdrivercd.utility import env_bool
from . import ScrewdriverTestCase

class UtilityTestCase(ScrewdriverTestCase):

    def setUp(self):
        super().setUp()
        self.testfiles = [
            f'{self.tempdir.name}/testfile',
            f'{self.tempdir.name}/.testfile',
        ]

    def _create_testfiles(self):
        for filename in self.testfiles:
            pathlib.Path(filename).touch()
            self.assertTrue(os.path.exists(filename))

    def test_clean_directory(self):
        self._create_testfiles()
        clean_directory(self.tempdir.name)
        self.assertFalse(os.path.exists(f'{self.tempdir.name}/testfile'))
        self.assertTrue(os.path.exists(f'{self.tempdir.name}/.testfile'))

    def test_copy_contents(self):
        os.makedirs('source')
        os.makedirs('dest')
        pathlib.Path(f'source/testfile').touch()
        pathlib.Path(f'source/.testfile').touch()

        self.assertFalse(os.path.exists(f'dest/testfile'))
        self.assertFalse(os.path.exists(f'dest/.testfile'))
        copy_contents('source', 'dest')
        self.assertTrue(os.path.exists(f'dest/testfile'))
        self.assertTrue(os.path.exists(f'dest/.testfile'))

    def test_copy_contents__skip_dotfiles(self):
        os.makedirs('source')
        os.makedirs('dest')
        pathlib.Path(f'source/testfile').touch()
        pathlib.Path(f'source/.testfile').touch()

        self.assertFalse(os.path.exists(f'dest/testfile'))
        self.assertFalse(os.path.exists(f'dest/.testfile'))
        copy_contents('source', 'dest', skip_dotfiles=True)
        self.assertTrue(os.path.exists(f'dest/testfile'))
        self.assertFalse(os.path.exists(f'dest/.testfile'))



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
screwdrivercd package functions
"""
import logging
import os
import subprocess  # nosec
import sys
import tarfile
import tempfile
import zipfile

from typing import Any, Dict

from pyroma.projectdata import run_setup, FakeContext, SetupMonkey
from .exceptions import PackageParseError


logger = logging.getLogger(__name__)


def run_setup_command(command, cli_args=None, stderr_log=None) -> bytes:
    """
    Run the setup command and return the results
    """
    command = [sys.executable, 'setup.py', command]
    if cli_args:
        command += cli_args
    if not stderr_log:  # pragma: no cover
        stderr_log = f'{os.environ.get("SD_ARTIFACTS_DIR", ".")}/logs/artifactory_publish.stderr.log'
    if not os.path.exists(os.path.dirname(stderr_log)):
        os.makedirs(os.path.dirname(stderr_log))
    with open(stderr_log, 'ab') as stderr:
        try:
            result = subprocess.check_output(command, stderr=stderr)  # nosec
        except subprocess.CalledProcessError as error:  # pragma: no cover
            if error.stdout:
                logger.error(f'{error.stdout}')
            raise error
    return result


def setup_query(command, cli_args=None) -> str:
    """
    Run a setup command and return the last line from stdout
    """
    return run_setup_command(command, cli_args=cli_args).decode(errors='ignore').strip().split(os.linesep)[-1].strip()


class PackageMetadata():
    """
    Python package metadata parser
    """
    used_setuptools = False
    _get_package_data_run = False
    metadata: Dict[str, Any] = {}
    options: Dict[str, Any] = {}

    def __init__(self, path=None):
        """
        Parse a python package

        Parameters
        ----------
        path: str, optional
            The path to the package, this can be a directory containing a package or a package archive file
        """
        super().__init__()
        if path:
            self.path = os.path.abspath(path)
        else:  # pragma: no cover
            self.path = os.path.abspath(os.getcwd())
        if os.path.isdir(self.path):  # pragma: no cover
            self.get_package_data()
        else:
            self.extract_package_data()

    def extract_package_data(self, path=None):
        """
        Extract a package archive and update the attributes of this class with it's metadata values.
        """
        # pylint: disable=W0612
        if not path:  # pragma: no cover
            path = self.path
        path = os.path.abspath(path)
        basename, extension = os.path.splitext(path)
        extension = extension.strip('.')
        if basename.endswith('.tar'):
            basename, junk = os.path.splitext(basename)

        basename = os.path.basename(basename)
        with tempfile.TemporaryDirectory() as tempdir:
            destdir = os.path.join(tempdir, basename)
            if extension in ['bz2', 'gz', 'tar', 'tb2', 'tgz']:
                logger.debug(f'Extracting tarfile {path} to {tempdir}')
                tarfile.open(name=path, mode='r:*').extractall(tempdir)
            elif extension in ['zip', 'egg']:
                logger.debug(f'Extracting zipfile {path} to {tempdir}')
                zipfile.ZipFile(path, mode='r').extractall(tempdir)
            else:  # pragma: no cover
                raise PackageParseError(f'Unknown package type {extension}')

            self.get_package_data(path=destdir)

    def get_package_data(self, path=None):
        """
        Updates the attributes of this class with values from the package.
        """
        # Run the imported setup to get the metadata.
        if not path:  # pragma: no cover
            path = self.path

        with FakeContext(path):
            with SetupMonkey() as sm:
                try:
                    distro = run_setup('setup.py', stop_after='config')
                except ImportError:  # pragma: no cover
                    raise PackageParseError('Unable to parse the package setup.py')

                self.used_setuptools = sm.used_setuptools
                for attrib in ['install_requires', 'metadata', 'namespace_packages', 'packages', 'python_requires', 'setup_requires']:
                    if not hasattr(distro, attrib):  # pragma: no cover
                        logging.debug(f'{attrib} is missing from distro object')
                        setattr(self, attrib, None)
                        continue

                    src_attrib = getattr(distro, attrib)
                    if not hasattr(src_attrib, '__dict__'):
                        setattr(self, attrib, src_attrib)
                        continue

                    dest_attrib_dict = {}
                    setattr(self, attrib, dest_attrib_dict)
                    src_attrib_dict = src_attrib.__dict__
                    for key, value in src_attrib_dict.items():
                        if key[0] == '_' or callable(value):  # pragma: no cover
                            continue
                        if not value:
                            dest_attrib_dict[key] = value
                            continue
                        dest_attrib_dict[key] = value

                self.command_options = distro.command_options
                self.options = self.command_options.get('options', {})
                for key, value in self.command_options.get('options', {}).items():
                    if isinstance(value, tuple):
                        if len(value) == 1:  # pragma: no cover
                            self.options[key] = None
                        elif len(value) == 2:
                            self.options[key] = value[1]
                        else:  # pragma: no cover
                            self.options[key] = value[1:]
                    else:  # pragma: no cover
                        self.options[key] = value

                for option_item in ['install_requires', 'namespace_packages', 'packages', 'python_requires', 'setup_requires']:
                    value = self.options.get(option_item, None)
                    if not value:
                        continue
                    if not isinstance(value, list):
                        self.options[option_item] = []
                        for item in [_.strip() for _ in value.split(os.linesep)]:
                            if item:
                                self.options[option_item].append(item)
                    if not hasattr(self, option_item):  # pragma: no cover
                        setattr(self, option_item, self.options[option_item])

                for option_item in ['include_package_data', 'zip_safe']:
                    value = self.options.get(option_item, None)
                    if not value:
                        continue
                    if value.lower() in ['true', '1', 'on']:  # pragma: no cover
                        self.options[option_item] = True
                    else:  # pragma: no cover
                        self.options[option_item] = False


def package_srcdir() -> str:
    # Package directory is being specifically set using env variables
    env_package_dir = os.environ.get('PACKAGE_DIR', '')
    if not env_package_dir:
        env_package_dir = os.environ.get('PACKAGE_DIRECTORY', '')

    if env_package_dir:
        return env_package_dir

    package_metadata = PackageMetadata()
    package_dir = package_metadata.options.get('package_dir', '').strip().lstrip('=')

    # Package directory specified in the package
    if package_dir:
        return package_dir

    package_name = package_metadata.metadata['name']

    # Fallback to looking for the package in a directory matching the package name
    if os.path.exists(package_name):
        return package_name

    for entry in os.listdir('.'):
        if entry.lower() == package_name.lower():
            return entry

    for entry in os.listdir('.'):
        if entry.lower().replace('-', '_') == package_name.lower().replace('-', '_'):
            return entry

    return ''



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Command line `sdv4_install_deps` utility.

The main() function of this utility provides a command line interface to install Operating system packages based on
the configuration in the `pyproject.toml` file.
"""
import json
import logging
import os
import sys

from termcolor import colored

from ..utility import create_artifact_directory
from .installer import Configuration
from .installers import install_plugins

LOG_NAME = 'platform_installdeps' if __name__ == '__main__' else __name__
LOG = logging.getLogger(LOG_NAME)


def main():  # pragma: no cover
    """
    Run all installers from the command line
    """
    loglevel = logging.WARNING
    if os.environ.get('INSTALLDEPS_DEBUG', 'false').lower() in ['true', '1', 'on']:
        loglevel = logging.DEBUG
    logging.basicConfig(level=loglevel)

    # if os.path.exists(LEGACY_CONFIG_FILE):
    #     print('Legacy configuration file exists, using the legacy installer', flush=True)
    #     return legacy_main()

    config = Configuration().configuration
    installer_order = config.get('install', None)
    fail_on_error = config.get('fail_on_error', False)
    if not installer_order:
        installer_order = install_plugins.keys()

    installed = {}
    for installer_name in installer_order:
        try:
            installer_class = install_plugins[installer_name]
        except KeyError:
            # No such installer class
            continue
        LOG.debug(f'Seeing if the {installer_name} tool is supported')
        installer_instance = installer_class()
        installer_instance.exit_on_missing = fail_on_error
        if not installer_instance.is_supported:
            continue
        if not installer_instance.has_dependencies:
            continue

        if installer_instance.plugin_configuration:
            print(f'Running {installer_name} installer ', flush=True)
            result = installer_instance.install_dependencies()
            if result:
                print(colored(f'Installed {len(result)} package{"s" if len(result) > 1 else ""}', 'green'), flush=True)
            else:
                print('No packages installed', flush=True)
            installed[installer_name] = result
            print('')

    # Make sure the report directory exists
    artifacts_dir = os.environ.get('SD_ARTIFACTS_DIR', '')
    report_dir = os.path.join(artifacts_dir, 'reports/installdeps')
    report_filename = os.path.join(report_dir, 'installdeps.json')
    create_artifact_directory(report_dir)

    with open(report_filename, 'w') as fh:
        json.dump(installed, fh)
    return 0


if __name__ == '__main__':
    sys.exit(main())
